import hashlib

from unittest import SkipTest

from django.test import TestCase

from hidp.accounts import tokens
from hidp.accounts.email_change import Recipient
from hidp.test.factories.user_factories import EmailChangeRequestFactory, UserFactory


class TestBaseEmailVerificationTokenGenerator(TestCase):
    token_generator = NotImplemented

    @classmethod
    def setUpClass(cls):
        if cls is TestBaseEmailVerificationTokenGenerator:
            raise SkipTest("Skipping abstract base class")
        super().setUpClass()

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()
        cls.token = cls.token_generator.make_token(cls.user)

    def test_check_invalid_token(self):
        """An invalid token returns none."""
        self.assertIsNone(self.token_generator.check_token("garbage-in:garbage-out"))

    def test_check_valid_token(self):
        """A valid token returns the MD5 sum of the email address."""
        self.assertEqual(
            self.token_generator.check_token(self.token),
            hashlib.md5(
                self.user.email.encode(),
                usedforsecurity=False,
            ).hexdigest(),
        )


class TestEmailVerificationRequestTokenGenerator(
    TestBaseEmailVerificationTokenGenerator
):
    token_generator = tokens.email_verification_request_token_generator


class TestEmailVerificationTokenGenerator(TestBaseEmailVerificationTokenGenerator):
    token_generator = tokens.email_verification_token_generator


class TestEmailChangeTokenGenerator(TestCase):
    token_generator = tokens.email_change_token_generator

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()
        cls.email_change_request = EmailChangeRequestFactory(user=cls.user)
        cls.current_token = cls.token_generator.make_token(
            str(cls.email_change_request.pk), Recipient.CURRENT_EMAIL
        )
        cls.proposed_token = cls.token_generator.make_token(
            str(cls.email_change_request.pk), Recipient.PROPOSED_EMAIL
        )

    def test_make_token_invalid_recipient(self):
        """An invalid recipient raises a ValueError."""
        with self.assertRaises(ValueError):
            self.token_generator.make_token(
                str(self.email_change_request.pk), "invalid-recipient"
            )

    def test_check_invalid_token(self):
        """An invalid token returns none."""
        self.assertIsNone(self.token_generator.check_token("garbage-in:garbage-out"))

    def test_check_valid_token(self):
        """A valid token returns the token object, containing uuid and recipient."""
        self.assertEqual(
            self.token_generator.check_token(self.current_token),
            {
                "uuid": str(self.email_change_request.pk),
                "recipient": Recipient.CURRENT_EMAIL,
            },
        )

        self.assertEqual(
            self.token_generator.check_token(self.proposed_token),
            {
                "uuid": str(self.email_change_request.pk),
                "recipient": Recipient.PROPOSED_EMAIL,
            },
        )



from django.test import SimpleTestCase

from hidp.federated.providers.microsoft import MicrosoftOIDCClient


class TestMicrosoftOIDCClient(SimpleTestCase):
    def test_initialize(self):
        """The Microsoft OIDC client can be initialized."""
        client = MicrosoftOIDCClient(
            client_id="test",
        )
        self.assertEqual(client.client_id, "test")
        self.assertEqual(client.callback_base_url, None)

    def test_get_issuer(self):
        """The issuer URL is formatted with the tenant ID."""
        client = MicrosoftOIDCClient(client_id="test")
        issuer = client.get_issuer(claims={"tid": "example"})
        self.assertEqual(issuer, "https://login.microsoftonline.com/example/v2.0")

        with self.subTest("Missing tenant ID"):
            # Returns the unformatted issuer URL if the tenant ID is missing.
            issuer = client.get_issuer(claims={})
            self.assertEqual(issuer, client.issuer)



from urllib.parse import urlencode

from django_otp import user_has_device

from django.shortcuts import redirect
from django.urls import reverse


class OTPMiddlewareBase:
    """
    Base class for OTP middleware.

    This class provides a base implementation for OTP middleware. It provides a
    ``process_view`` method that checks whether a request needs to verify OTP and
    redirects to the OTP verification view if necessary. The conditions on when to
    require verification can be implemented by overriding the
    ``user_needs_verification`` method in a subclass. For more complex verification
    logic, you can override the ``request_needs_verification`` and
    ``view_func_needs_verification`` methods.

    Views can be marked as exempt from OTP verification by using the ``otp_exempt``
    decorator.

    Middleware implementations should be placed after the authentication middleware
    and ``django_otp.middleware.OTPMiddleware``. If ``request_needs_verification``,
    it will redirect users to the OTP verification view if they have a configured OTP
    device, or else to the OTP setup view.
    """

    def __new__(cls, *args, **kwargs):
        if cls is OTPMiddlewareBase:
            raise TypeError(
                f"{cls.__name__} cannot be used directly, use one of the "
                "subclasses instead"
            )
        return super().__new__(cls)

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def _view_is_exempt(self, view_func):  # noqa: PLR6301
        """
        Check if a view function is exempt from OTP verification.

        A view is exempt if it has the `otp_exempt` attribute set to `True` by the
        `otp_exempt` decorator.
        """
        return getattr(view_func, "otp_exempt", False)

    def get_redirect_url(self, request):  # noqa: PLR6301
        """
        Return the URL to redirect to when OTP verification is required.

        If the user has an OTP device, they will be redirected to the OTP verification
        view. If they do not have an OTP device, they will be redirected to the OTP
        setup view.

        Args:
            request (``HttpRequest``): The request object.

        Returns:
            str: The URL to redirect to.
        """
        target = reverse(
            "hidp_otp:verify"
            if user_has_device(request.user)
            else "hidp_otp_management:setup"
        )
        params = {"next": request.get_full_path()}
        return f"{target}?{urlencode(params)}"

    def user_needs_verification(self, user):  # noqa: PLR6301
        """
        Check if a user needs to verify their OTP.

        By default, this will require all (authenticated and not verified) users to
        verify. Override this method to customize the verification logic.

        Args:
            user (``User``): The user to check.

        Returns:
            bool: Whether the user needs to verify their OTP.
        """
        return True

    def view_func_needs_verification(self, view_func):
        """
        Check whether a view function needs to verify OTP.

        Override this method if you need to customize the verification logic.

        Args:
            view_func (callable): The view function to check.

        Returns:
            bool: Whether the view function needs to verify OTP.
        """
        return not self._view_is_exempt(view_func)

    def request_needs_verification(self, request, view_func):
        """
        Check whether a request needs to verify OTP.

        The request needs to verify OTP if the user needs verification and the view
        function needs verification. The user never needs to verify if they are not
        authenticated or already verified.

        Args:
            request (``HttpRequest``): The request object.
            view_func (callable): The view function that will be called.

        Returns:
            bool: Whether the request requires the user to verify OTP.
        """
        return (
            self.view_func_needs_verification(view_func)
            and request.user.is_authenticated
            and not request.user.is_verified()
            and self.user_needs_verification(request.user)
        )

    def process_view(self, request, view_func, view_args, view_kwargs):
        """
        Process a view and check if OTP verification is required.

        This method is called by the middleware to check if the user needs to verify
        their OTP. If verification is required, the user will be redirected to the OTP
        verification view, or to the OTP setup view if they do not have an OTP device.

        Args:
            request (``HttpRequest``): The request object.
            view_func (callable): The view function that will be called.
            view_args (list): The positional arguments passed to the view function.
            view_kwargs (dict): The keyword arguments passed to the view function.

        Returns:
            ``None`` or ``HttpResponseRedirect``: HttpResponseRedirect if this policy
            requires the user to verify OTP, None otherwise.
        """
        if self.request_needs_verification(request, view_func):
            return redirect(self.get_redirect_url(request))

        return None


class OTPRequiredMiddleware(OTPMiddlewareBase):
    """
    Middleware that requires all users to verify their OTP.

    This middleware should be placed after the authentication middleware and
    django_otp.middleware.OTPMiddleware. It will redirect users to the OTP
    verification view if they are authenticated and have not yet verified their OTP,
    or to the OTP setup view if they have not yet configured an OTP device.
    """


class OTPVerificationRequiredIfConfiguredMiddleware(OTPMiddlewareBase):
    """
    Middleware that requires users to verify their OTP if they have OTP configured.

    This middleware should be placed after the authentication middleware and
    django_otp.middleware.OTPMiddleware. It will redirect users to the OTP
    verification view if they are authenticated, have a configured OTP device,
    but have not yet verified their OTP, or to the OTP setup view if they have not
    yet configured an OTP device.
    """

    def user_needs_verification(self, user):  # noqa: PLR6301
        """
        Check if a user needs to verify their OTP.

        A user needs to verify their OTP if they are authenticated, have an OTP
        device, and have not yet verified their OTP.
        """
        return user_has_device(user)


class OTPSetupRequiredIfStaffUserMiddleware(OTPMiddlewareBase):
    """
    Middleware that requires staff users to configure and verify their OTP.

    This middleware should be placed after the authentication middleware and
    django_otp.middleware.OTPMiddleware. It will redirect staff users to the OTP
    verification view if they are authenticated and are staff, even if they do not
    have an OTP device configured. If they don't have an OTP device configured, they
    will be redirected to the OTP setup view.
    """

    def user_needs_verification(self, user):  # noqa: PLR6301
        """
        Check if a user needs to verify their OTP.

        A user needs to verify their OTP if they are authenticated, are staff, and
        have not yet verified their OTP.
        """
        return user.is_staff



"""
OIDC client URLs.

Provides the URL patterns for OIDC client authentication and registration.

Include this module in the root URL configuration:

    from hidp.federated import oidc_client_urls

    urlpatterns = [
        path("login/oidc/", include(oidc_client_urls)),
    ]

This module also defines the namespace `hidp_oidc_client` for these URLs.

Include this namespace when reversing URLs, for example:

    reverse("hidp_oidc_client:authenticate", kwargs={"provider_key": "example"})
"""

from django.urls import path

from . import views

app_name = "hidp_oidc_client"

urlpatterns = [
    path(
        "authenticate/<slug:provider_key>/",
        views.OIDCAuthenticationRequestView.as_view(),
        name="authenticate",
    ),
    path(
        "reauthenticate/<slug:provider_key>/",
        views.OIDCAuthenticationRequestView.as_view(
            extra_authentication_request_params={
                "prompt": "login",
                "max_age": 0,
            }
        ),
        name="reauthenticate",
    ),
    path(
        "callback/<slug:provider_key>/",
        views.OIDCAuthenticationCallbackView.as_view(),
        name="callback",
    ),
    path(
        "register/",
        views.OIDCRegistrationView.as_view(),
        name="register",
    ),
    path(
        "login/",
        views.OIDCLoginView.as_view(),
        name="login",
    ),
]



from django import forms
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from ..accounts.forms import TermsOfServiceMixin
from .models import OpenIdConnection

UserModel = get_user_model()


class OIDCRegistrationForm(TermsOfServiceMixin, forms.ModelForm):
    """Create a user and OpenIDConnection from OIDC claims and user info."""

    template_name = "hidp/federated/forms/registration_form.html"

    # Fields
    agreed_to_tos = TermsOfServiceMixin.create_agreed_to_tos_field()

    def __init__(self, *, provider_key, claims, user_info, **kwargs):
        self.provider_key = provider_key
        self.claims = claims
        self.user_info = user_info

        # Populate the form using the OIDC claims and user info.
        oidc_data = claims | user_info
        initial_data = {
            "email": oidc_data.get("email"),
            "first_name": oidc_data.get("given_name"),
            "last_name": oidc_data.get("family_name"),
        } | kwargs.pop("initial", {})

        super().__init__(initial=initial_data, **kwargs)

        # Disable the email field to prevent the user from changing it.
        self.fields["email"].disabled = True

        # If the given name and family name are not provided, remove the fields.
        # They will be asked for during email verification.
        if not initial_data["first_name"] and not initial_data["last_name"]:
            self.fields.pop("first_name")
            self.fields.pop("last_name")
        else:
            self.fields["first_name"].required = True
            self.fields["last_name"].required = True

    class Meta:
        model = UserModel
        fields = [
            "email",
            "first_name",
            "last_name",
        ]

    @transaction.atomic
    def save(self, *, commit=True):
        user = super().save(commit=False)
        user.set_unusable_password()
        self.set_agreed_to_tos(user)
        user.connection = OpenIdConnection(
            user=user,
            provider_key=self.provider_key,
            issuer_claim=self.claims["iss"],
            subject_claim=self.claims["sub"],
        )
        if commit:
            user.save()
            user.connection.save()
        return user


class OIDCAccountLinkForm(forms.ModelForm):
    """Link an existing user to an OpenIDConnection."""

    template_name = "hidp/federated/forms/account_link_form.html"

    # Fields
    allow_link = forms.BooleanField(
        label=_("Yes, I want to link this account"),
        required=True,
        initial=True,
        widget=forms.HiddenInput(),
    )

    def __init__(self, *, user, provider_key, claims, **kwargs):
        self.user = user
        self.provider_key = provider_key
        self.claims = claims
        super().__init__(**kwargs)

    class Meta:
        model = OpenIdConnection
        fields = []

    @transaction.atomic
    def save(self, *, commit=True):
        self.instance = OpenIdConnection(
            user=self.user,
            provider_key=self.provider_key,
            issuer_claim=self.claims["iss"],
            subject_claim=self.claims["sub"],
        )
        if commit:
            self.instance.save()
        return self.instance


class OIDCAccountUnlinkForm(forms.Form):
    """Unlink an OpenIDConnection from a user."""

    template_name = "hidp/federated/forms/account_unlink_form.html"

    # Fields
    allow_unlink = forms.BooleanField(
        label=_("Yes, I want to unlink this account"),
        required=True,
        initial=True,
        widget=forms.HiddenInput(),
    )

    def __init__(self, *, user, provider_key, **kwargs):
        self.user = user
        self.provider_key = provider_key
        super().__init__(**kwargs)

    def clean(self):
        cleaned_data = super().clean()
        # Do not allow the user to unlink the only available login method.
        can_unlink = (
            self.user.has_usable_password()
            or self.user.openid_connections.exclude(
                provider_key=self.provider_key
            ).exists()
        )
        if not can_unlink:
            raise ValidationError(_("You cannot unlink your only way to sign in."))
        return cleaned_data



from django.test import TestCase, override_settings


@override_settings(ROOT_URLCONF="tests.unit_tests.test_rate_limit.urls")
class TestRateLimitMiddleware(TestCase):
    def test_rate_limited_view(self):
        response = self.client.get("/rate_limited_view/")
        self.assertEqual(response.status_code, 429)
        self.assertEqual(
            response.content.decode(),
            (
                "Sorry, you have made too many requests to the server."
                " Please try again later."
            ),
        )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Hello, ID Please"
project_copyright = "2024, Jaap Roes, Dennis Bunskoek, Thomas Kalverda, Wouter de Vries"
author = "Jaap Roes, Dennis Bunskoek, Thomas Kalverda, Wouter de Vries"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["myst_parser", "sphinx.ext.napoleon", "sphinx.ext.autodoc"]
myst_enable_extensions = ["colon_fence", "attrs_inline", "deflist"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autodoc_mock_imports = ["django"]

myst_heading_anchors = 3

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



import io

from http import HTTPStatus
from unittest import mock

from django.contrib import auth
from django.core import mail
from django.core.management import call_command
from django.test import TestCase
from django.urls import reverse
from django.utils import timezone

from hidp.accounts import tokens
from hidp.accounts.email_verification import remove_stale_unverified_accounts
from hidp.accounts.forms import EmailVerificationForm
from hidp.test.factories import user_factories

UserModel = auth.get_user_model()


class TestEmailVerificationRequiredView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.url = reverse(
            "hidp_accounts:email_verification_required",
            kwargs={
                "token": tokens.email_verification_request_token_generator.make_token(
                    cls.user
                )
            },
        )

    def _assert_response(self, response, *, validlink=True):
        """Convenience method to assert the response."""
        self.assertEqual(response.status_code, HTTPStatus.OK)
        if validlink:
            self.assertTemplateUsed(
                response, "hidp/accounts/verification/email_verification_required.html"
            )
        else:
            self.assertTemplateUsed(
                response,
                "hidp/accounts/verification/email_verification_required_invalid_link.html",
            )

    def test_valid_get(self):
        """Works when the token is considered valid."""
        self._assert_response(self.client.get(self.url, follow=True))

    def test_get_invalid_token(self):
        """Invalid token."""
        response = self.client.get(
            reverse(
                "hidp_accounts:email_verification_required",
                kwargs={"token": "invalid-value:invalid-signature"},
            ),
            follow=True,
        )
        self._assert_response(response, validlink=False)

    def test_no_token_in_session(self):
        """Placeholder token, no token in session."""
        response = self.client.get(
            reverse(
                "hidp_accounts:email_verification_required",
                kwargs={"token": "email"},
            ),
            follow=True,
        )
        self._assert_response(response, validlink=False)

    def test_post(self):
        """Send the verification email."""
        # Get the page first, to populate the session
        response = self.client.get(self.url, follow=True)
        # Post to the redirected URL
        with (
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_body.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_body.html"),
        ):  # fmt: skip
            self.client.post(response.redirect_chain[-1][0], follow=True)
        # Verification email sent
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(
            message.subject,
            "Verify your email address",
        )

    def test_post_invalid_token(self):
        """Does not send the verification email when the token is invalid."""
        # Get the page first, to populate the session
        response = self.client.get(
            reverse(
                "hidp_accounts:email_verification_required",
                kwargs={"token": "invalid-value:invalid-signature"},
            ),
            follow=True,
        )
        # Post to the redirected URL
        self.client.post(response.redirect_chain[-1][0], follow=True)
        # Verification email not sent
        self.assertEqual(len(mail.outbox), 0)
        self._assert_response(response, validlink=False)


class TestEmailVerificationView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.url = reverse(
            "hidp_accounts:verify_email",
            kwargs={
                "token": tokens.email_verification_token_generator.make_token(cls.user)
            },
        )

    def _assert_response(self, response, *, validlink=True):
        """Convenience method to assert the response."""
        self.assertEqual(response.status_code, HTTPStatus.OK)
        if validlink:
            self.assertTemplateUsed(
                response, "hidp/accounts/verification/verify_email.html"
            )
        else:
            self.assertTemplateUsed(
                response, "hidp/accounts/verification/verify_email_invalid_link.html"
            )

    def test_valid_get(self):
        """Works when the token is considered valid."""
        self._assert_response(self.client.get(self.url, follow=True))

    def test_get_invalid_token(self):
        """Invalid token."""
        response = self.client.get(
            reverse(
                "hidp_accounts:verify_email",
                kwargs={"token": "invalid-value:invalid-signature"},
            ),
            follow=True,
        )
        self._assert_response(response, validlink=False)

    def test_no_token_in_session(self):
        """Placeholder token, no token in session."""
        response = self.client.get(
            reverse(
                "hidp_accounts:verify_email",
                kwargs={"token": "email"},
            ),
            follow=True,
        )
        self._assert_response(response, validlink=False)

    def test_inactive_user(self):
        """Inactive user."""
        self.user.is_active = False
        self.user.save()
        response = self.client.get(self.url, follow=True)
        self._assert_response(response, validlink=False)

    def test_already_verified_user(self):
        """Already verified user."""
        self.user.email_verified = timezone.now()
        self.user.save()
        response = self.client.get(self.url, follow=True)
        self._assert_response(response, validlink=False)

    def test_post(self):
        """Update the user's email_verified field."""
        # Get the page first, to populate the session
        response = self.client.get(self.url, follow=True)
        # Post to the redirected URL
        response = self.client.post(response.redirect_chain[-1][0], follow=True)
        self.user.refresh_from_db()
        self.assertIsNotNone(
            self.user.email_verified, msg="Expected email to be verified."
        )
        self.assertAlmostEqual(
            self.user.email_verified,
            timezone.now(),
            delta=timezone.timedelta(seconds=5),
        )
        self.assertURLEqual(
            response.redirect_chain[-1][0],
            reverse("hidp_accounts:email_verification_complete"),
        )

    def test_post_invalid_token(self):
        # Get the page first, to populate the session
        response = self.client.get(
            reverse(
                "hidp_accounts:verify_email",
                kwargs={"token": "invalid-value:invalid-signature"},
            ),
            follow=True,
        )
        # Post to the redirected URL
        self.client.post(response.redirect_chain[-1][0], follow=True)
        # User's email_verified field not updated
        self.user.refresh_from_db()
        self.assertIsNone(
            self.user.email_verified, msg="Expected email to not be verified."
        )
        self._assert_response(response, validlink=False)


class TestEmailVerificationForm(TestCase):
    def test_form_for_user_no_name(self):
        user = user_factories.UserFactory(first_name="", last_name="")
        form = EmailVerificationForm(instance=user)
        self.assertEqual(form.initial["first_name"], "")
        self.assertEqual(form.initial["last_name"], "")
        self.assertTrue(form.fields["first_name"].required)
        self.assertTrue(form.fields["last_name"].required)

    def test_form_for_user_with_partial_name(self):
        user = user_factories.UserFactory(first_name="", last_name="White")
        form = EmailVerificationForm(instance=user)
        self.assertEqual(form.initial["first_name"], "")
        self.assertEqual(form.initial["last_name"], "White")
        self.assertTrue(form.fields["first_name"].required)
        self.assertTrue(form.fields["last_name"].required)

    def test_form_for_user_with_name(self):
        """
        Form should not have first and last name fields.

        The user has probably been created using an OIDC connection, and the given name
        and family name were provided by the OIDC provider.
        """
        user = user_factories.UserFactory(first_name="Walter", last_name="White")
        form = EmailVerificationForm(instance=user)
        self.assertNotIn("first_name", form.fields)
        self.assertNotIn("last_name", form.fields)

    def test_updates_name_fields(self):
        user = user_factories.UserFactory(first_name="", last_name="")
        form = EmailVerificationForm(
            instance=user,
            data={
                "first_name": "Walter",
                "last_name": "White",
            },
        )
        self.assertTrue(form.is_valid())
        form.save()
        user.refresh_from_db()
        self.assertEqual(user.first_name, "Walter")
        self.assertEqual(user.last_name, "White")


class TestRemoveUnverifiedAccounts(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.stale_user = user_factories.UserFactory(
            email_verified=None,
            date_joined=timezone.now() - timezone.timedelta(days=91),
        )
        cls.unverified_user = user_factories.UserFactory(
            date_joined=timezone.now() - timezone.timedelta(days=31)
        )
        cls.verified_user = user_factories.VerifiedUserFactory()

    def test_remove_stale_unverified_accounts_dry_run(self):
        removed_accounts = remove_stale_unverified_accounts(dry_run=True)
        users_exist = (
            UserModel.objects.filter(pk=user.pk).exists()
            for user in (self.stale_user, self.unverified_user, self.verified_user)
        )
        self.assertEqual(
            removed_accounts,
            1,
            msg="Expected 1 account to be selected for removal.",
        )
        self.assertTrue(
            all(users_exist),
            msg="Expected all users to still exist after dry run.",
        )

    def test_remove_stale_unverified_accounts(self):
        removed_accounts = remove_stale_unverified_accounts()
        self.assertEqual(
            removed_accounts,
            1,
            msg="Expected 1 account to be selected for removal.",
        )
        self.assertFalse(
            UserModel.objects.filter(pk=self.stale_user.pk).exists(),
            msg="Expected stale user to be removed, joined more than 90 days ago.",
        )
        users_exist = (
            UserModel.objects.filter(pk=user.pk).exists()
            for user in (self.unverified_user, self.verified_user)
        )
        self.assertTrue(
            all(users_exist),
            msg="Expected non-stale users to still exist.",
        )

    def test_remove_stale_unverified_accounts_30_days(self):
        removed_accounts = remove_stale_unverified_accounts(days=30)
        self.assertEqual(removed_accounts, 2)
        self.assertFalse(
            UserModel.objects.filter(pk=self.stale_user.pk).exists(),
            msg="Expected user to be removed, joined more than 30 days ago.",
        )
        self.assertFalse(
            UserModel.objects.filter(pk=self.unverified_user.pk).exists(),
            msg="Expected user to be removed, joined more than 30 days ago.",
        )
        self.assertTrue(
            UserModel.objects.filter(pk=self.verified_user.pk).exists(),
            msg="Expected user to still exist, user is verified.",
        )

    @mock.patch(
        "hidp.accounts.management.commands.remove_stale_unverified_accounts.remove_stale_unverified_accounts",
        return_value=1,
    )
    def test_remove_stale_unverified_accounts_management_command_dry_run(
        self, mock_remove_stale_unverified_accounts
    ):
        stdout = io.StringIO()

        call_command("remove_stale_unverified_accounts", dry_run=True, stdout=stdout)
        self.assertIn(
            "Removing accounts that have not been verified within 90 days...",
            stdout.getvalue(),
        )
        self.assertIn("1 unverified account(s) would be removed.", stdout.getvalue())
        mock_remove_stale_unverified_accounts.assert_called_once_with(
            days=90, dry_run=True
        )

    @mock.patch(
        "hidp.accounts.management.commands.remove_stale_unverified_accounts.remove_stale_unverified_accounts",
        return_value=1,
    )
    def test_remove_stale_unverified_accounts_management_command(
        self, mock_remove_stale_unverified_accounts
    ):
        stdout = io.StringIO()

        call_command("remove_stale_unverified_accounts", stdout=stdout)
        self.assertIn(
            "Removing accounts that have not been verified within 90 days...",
            stdout.getvalue(),
        )
        self.assertIn(
            "Successfully removed 1 unverified account(s).", stdout.getvalue()
        )
        mock_remove_stale_unverified_accounts.assert_called_once_with(
            days=90, dry_run=False
        )

    @mock.patch(
        "hidp.accounts.management.commands.remove_stale_unverified_accounts.remove_stale_unverified_accounts",
        return_value=2,
    )
    def test_remove_stale_unverified_accounts_management_command_30_days(
        self, mock_remove_stale_unverified_accounts
    ):
        stdout = io.StringIO()

        call_command("remove_stale_unverified_accounts", days=30, stdout=stdout)
        self.assertIn(
            "Successfully removed 2 unverified account(s).", stdout.getvalue()
        )
        mock_remove_stale_unverified_accounts.assert_called_once_with(
            days=30, dry_run=False
        )



from rest_framework import serializers

from django.contrib.auth import get_user_model

UserModel = get_user_model()


class UserSerializer(serializers.ModelSerializer):
    first_name = serializers.CharField(required=True)
    last_name = serializers.CharField(required=True)

    class Meta:
        model = UserModel
        fields = ["first_name", "last_name"]



from http import HTTPStatus
from unittest import mock

from django.core import mail
from django.test import TestCase
from django.urls import reverse

from hidp.accounts import forms, mailers
from hidp.test.factories import user_factories


class TestPasswordResetFlow(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()

    def test_get_password_reset_request(self):
        """Render the password reset request form."""
        response = self.client.get(reverse("hidp_accounts:password_reset_request"))
        self.assertTemplateUsed(
            response, "hidp/accounts/recovery/password_reset_request.html"
        )
        self.assertIsInstance(response.context["form"], forms.PasswordResetRequestForm)

    def test_non_user_request_password_reset_email(self):
        """A non-user cannot request a password reset email."""
        response = self.client.post(
            reverse("hidp_accounts:password_reset_request"),
            {"email": "not-a-user@example.com"},
        )
        self.assertEqual(0, len(mail.outbox))
        # Even though the email was not sent, the user is redirected to
        # the success page to prevent email enumeration attacks.
        self.assertRedirects(
            response,
            reverse("hidp_accounts:password_reset_email_sent"),
            fetch_redirect_response=True,
        )

    def test_inactive_user_request_password_reset_email(self):
        """An inactive user cannot request a password reset email."""
        self.user.is_active = False
        self.user.save()
        response = self.client.post(
            reverse("hidp_accounts:password_reset_request"),
            {"email": self.user.email},
        )
        self.assertEqual(0, len(mail.outbox))
        self.assertRedirects(
            response,
            reverse("hidp_accounts:password_reset_email_sent"),
            fetch_redirect_response=True,
        )

    def test_user_without_a_password_request_password_reset_email(self):
        self.user.set_unusable_password()
        self.user.save()
        """A user without a password receives the set password email."""
        with (
            self.assertTemplateUsed("hidp/accounts/recovery/email/set_password_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/recovery/email/set_password_body.txt"),
            self.assertTemplateUsed("hidp/accounts/recovery/email/set_password_body.html")
        ):  # fmt: skip
            response = self.client.post(
                reverse("hidp_accounts:password_reset_request"),
                {"email": self.user.email},
                follow=True,
            )
        self.assertEqual(1, len(mail.outbox))
        message = mail.outbox[0]
        self.assertEqual(message.to, [self.user.email])
        self.assertEqual(message.subject, "Set a password")
        self.assertIn(
            reverse("hidp_account_management:set_password"),
            message.body,
        )
        self.assertRedirects(
            response,
            reverse("hidp_accounts:password_reset_email_sent"),
        )
        self.assertTemplateUsed(
            response,
            "hidp/accounts/recovery/password_reset_email_sent.html",
        )

    def test_user_request_password_reset_email(self):
        """A user can request a password reset email."""
        with (
            self.assertTemplateUsed("hidp/accounts/recovery/email/password_reset_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/recovery/email/password_reset_body.txt"),
            self.assertTemplateUsed("hidp/accounts/recovery/email/password_reset_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                reverse("hidp_accounts:password_reset_request"),
                {"email": self.user.email},
                follow=True,
            )
        self.assertEqual(1, len(mail.outbox))
        message = mail.outbox[0]
        self.assertEqual(message.to, [self.user.email])
        self.assertEqual(message.subject, "Reset your password")
        self.assertRegex(
            message.body,
            # Matches the password reset URL:
            # http://testserver/recover/password/MDE5MTkyY2UtODE0Yy03NjNlLTlhMGUtMmM1ODk3MGNkYTFj/cced4c-9a0766ea185039a6d293ff660c04007e/
            r"http://testserver/recover/password/[0-9A-Za-z]+/[0-9a-z]+-[0-9a-f]+/",
        )
        self.assertRedirects(
            response,
            reverse("hidp_accounts:password_reset_email_sent"),
        )
        self.assertTemplateUsed(
            response,
            "hidp/accounts/recovery/password_reset_email_sent.html",
        )

    @mock.patch(
        "hidp.accounts.views.mailers.PasswordResetRequestMailer.send",
        side_effect=Exception,
    )
    def test_password_reset_email_error(self, mock_send):
        """Errors are logged when sending the password reset email."""
        # This is a fix for Django's CVE-2024-45231. The email backend
        # might raise an exception when sending an email, which could
        # be used to enumerate valid email addresses.
        with self.assertLogs("hidp.accounts.views", level="ERROR") as cm:
            self.client.post(
                reverse("hidp_accounts:password_reset_request"),
                {"email": self.user.email},
            )
        self.assertIn("Failed to send password (re)set email.", cm.records[0].msg)
        self.assertEqual(0, len(mail.outbox))

    def test_get_password_reset_url(self):
        """Render the password reset form."""
        password_reset_url = mailers.PasswordResetRequestMailer(
            user=self.user,
            base_url="https://testserver",
        ).get_password_reset_url()
        response = self.client.get(password_reset_url, follow=True)
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(response, "hidp/accounts/recovery/password_reset.html")
        self.assertIsInstance(response.context["form"], forms.PasswordResetForm)

    def test_post_password_reset_url(self):
        """Reset the user's password."""
        password_reset_url = mailers.PasswordResetRequestMailer(
            user=self.user,
            base_url="https://testserver",
        ).get_password_reset_url()
        # Need to get the password reset form first to populate a session value.
        response = self.client.get(
            password_reset_url,
            follow=True,
        )
        self.assertEqual(response.status_code, HTTPStatus.OK)
        with (
            self.assertTemplateUsed("hidp/accounts/management/email/password_changed_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/password_changed_body.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/password_changed_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                # There is a redirect to remove the token from the URL.
                # The final destination is the URL we need to POST to.
                response.redirect_chain[-1][0],
                {
                    "new_password1": "newpassword",
                    "new_password2": "newpassword",
                },
                follow=True,
            )
        self.user.refresh_from_db()
        self.assertTrue(self.user.check_password("newpassword"))
        self.assertRedirects(
            response,
            reverse("hidp_accounts:password_reset_complete"),
        )
        self.assertTemplateUsed(
            response,
            "hidp/accounts/recovery/password_reset_complete.html",
        )

        # Changed password mail should be sent
        self.assertEqual(1, len(mail.outbox))
        message = mail.outbox[0]
        self.assertEqual(message.to, [self.user.email])
        self.assertEqual("Your password has been changed", message.subject)
        self.assertIn(
            reverse("hidp_accounts:password_reset_request"),
            message.body,
        )

        with self.subTest("The password reset URL is invalid after use."):
            response = self.client.get(password_reset_url, follow=True)
            self.assertTemplateUsed(
                response,
                "hidp/accounts/recovery/password_reset_invalid_link.html",
            )



from django.test import SimpleTestCase

from hidp.federated.providers.google import GoogleOIDCClient


class TestGoogleOIDCClient(SimpleTestCase):
    def test_requires_client_secret(self):
        """The Google OIDC client requires a client secret."""
        with self.assertRaisesMessage(
            TypeError,
            "GoogleOIDCClient.__init__() missing 1 required"
            " keyword-only argument: 'client_secret'",
        ):
            GoogleOIDCClient(client_id="test")

    def test_initialize(self):
        """The Google OIDC client can be initialized."""
        client = GoogleOIDCClient(
            client_id="test",
            client_secret="secret",
            callback_base_url="https://example.com/",
        )
        self.assertEqual(client.client_id, "test")
        self.assertEqual(client.callback_base_url, "https://example.com/")



import hashlib

from datetime import timedelta

from django.core import signing

from .email_change import Recipient


class BaseTokenGenerator:
    key_salt = NotImplemented
    token_timeout = NotImplemented

    def __init__(self):
        self._signer = self._get_signer()

    def _get_signer(self):
        return signing.TimestampSigner(algorithm="sha256", salt=self.key_salt)

    def make_token(self, value):
        """
        Generate an expiring token based on a value.

        Args:
            value:
                The value to use to generate the token.

        Returns:
            str: The generated token.
        """
        if value is None:
            # The value should not be `None` because it would be impossible to verify as
            # `None` is returned when the token is invalid or expired.
            raise ValueError("The value cannot be None.")
        return self._signer.sign_object(value)

    def check_token(self, token):
        """
        Verify a token.

        Args:
            token (str): The token to verify.

        Returns:
            value | None:
                The value used to generate the token,
                or `None` if the token is invalid or expired.
        """
        try:
            return self._signer.unsign_object(token, max_age=self.token_timeout)
        except signing.BadSignature:
            return None


class BaseEmailVerificationTokenGenerator(BaseTokenGenerator):
    def make_token(self, user):
        """
        Generate an expiring token based on the user's email address.

        Args:
            user (User): The user to generate the token for.

        Returns:
            str: The generated token.
        """
        # Create a MD5 hash of the user's email address.
        # MD5 is used here to create a fixed length hash that is not reversible,
        # and can be used for easy and cheap database lookups.
        # The hash is then signed to prevent tampering.
        return super().make_token(
            hashlib.md5(
                user.email.encode(),
                usedforsecurity=False,
            ).hexdigest()
        )


class EmailVerificationRequestTokenGenerator(BaseEmailVerificationTokenGenerator):
    """Token to request a new email verification link."""

    key_salt = "email-verification-request"
    token_timeout = timedelta(hours=1).total_seconds()


email_verification_request_token_generator = EmailVerificationRequestTokenGenerator()


class EmailVerificationTokenGenerator(BaseEmailVerificationTokenGenerator):
    """Token to verify a user's email address."""

    key_salt = "email-verification"
    token_timeout = timedelta(days=1).total_seconds()


email_verification_token_generator = EmailVerificationTokenGenerator()


class EmailChangeTokenGenerator(BaseTokenGenerator):
    """Token to confirm an email change request."""

    key_salt = "email-change"
    token_timeout = timedelta(days=1).total_seconds()

    def make_token(self, email_change_request_uuid, recipient):
        """Generate a token based on the email change request and recipient."""
        if recipient not in Recipient.__members__.values():
            raise ValueError(f"Invalid recipient: {recipient}")

        return super().make_token(
            {
                "uuid": str(email_change_request_uuid),
                "recipient": recipient,
            }
        )


email_change_token_generator = EmailChangeTokenGenerator()



# Generated by Django 5.1.1 on 2024-10-03 10:24

import django.db.models.deletion
import hidp.compat.uuid7
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='EmailChangeRequest',
            fields=[
                ('id', models.UUIDField(default=hidp.compat.uuid7.uuid7, editable=False, primary_key=True, serialize=False)),
                ('proposed_email', models.EmailField(max_length=254, verbose_name='proposed email address')),
                ('current_email', models.EmailField(max_length=254, verbose_name='current email address')),
                ('confirmed_by_proposed_email', models.BooleanField(default=False, verbose_name='confirmed by proposed email')),
                ('confirmed_by_current_email', models.BooleanField(default=False, verbose_name='confirmed by current email')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('last_modified', models.DateTimeField(auto_now=True, verbose_name='last modified')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='email_change_request', to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'verbose_name': 'email change request',
                'verbose_name_plural': 'email change requests',
            },
        ),
    ]



from unittest import TestCase

from django.http import HttpResponse

from hidp.otp.decorators import otp_exempt


@otp_exempt
def no_otp_required_for_this_view(request):
    return HttpResponse("No OTP required for this view.")


class TestOTPExemptDecorator(TestCase):
    def test_otp_exempt_decorator(self):
        self.assertTrue(
            hasattr(no_otp_required_for_this_view, "otp_exempt"),
            msg="Expected the view to have an `otp_exempt` attribute.",
        )
        self.assertTrue(
            no_otp_required_for_this_view.otp_exempt,
            msg="Expected the view to be exempt from OTP verification.",
        )



from unittest import TestCase

from hidp.accounts import mailers


class TestBaseMailer(TestCase):
    def test_default_context(self):
        mailer = mailers.BaseMailer(base_url="https://example.com/")
        self.assertEqual(
            mailer.get_context(),
            {
                "base_url": "https://example.com",
            },
        )

    def test_get_recipients(self):
        with self.assertRaises(NotImplementedError) as cm:
            mailers.BaseMailer(base_url="https://example.com/").get_recipients()
        self.assertEqual(
            cm.exception.args[0], "Method get_recipients must be implemented"
        )

    def test_send(self):
        with self.assertRaises(NotImplementedError) as cm:
            mailers.BaseMailer(base_url="https://example.com/").send()
        self.assertEqual(
            cm.exception.args[0], "Attribute subject_template_name must be set"
        )



from datetime import timedelta
from http import HTTPStatus

from django.core import mail
from django.test import TestCase
from django.urls import reverse
from django.utils import timezone

from hidp.accounts.forms import EditUserForm
from hidp.test.factories import user_factories


class TestManageAccountView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.manage_account_url = reverse("hidp_account_management:manage_account")

    def test_login_required(self):
        """Anonymous users should be redirected to the login page."""
        response = self.client.get(self.manage_account_url)
        self.assertRedirects(
            response,
            f"{reverse('hidp_accounts:login')}?next={self.manage_account_url}",
        )

    def test_get(self):
        """The manage account page should be displayed for authenticated users."""
        self.client.force_login(self.user)
        response = self.client.get(self.manage_account_url)
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(
            response, "hidp/accounts/management/manage_account.html"
        )


class TestEditAccountView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.edit_account_url = reverse("hidp_account_management:edit_account")

    def test_login_required(self):
        """Anonymous users should be redirected to the login page."""
        response = self.client.get(self.edit_account_url)
        self.assertRedirects(
            response,
            f"{reverse('hidp_accounts:login')}?next={self.edit_account_url}",
        )

    def test_get(self):
        """The edit user page should be displayed for authenticated users."""
        self.client.force_login(self.user)
        response = self.client.get(self.edit_account_url)
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(response, "hidp/accounts/management/edit_account.html")
        self.assertIn("form", response.context)
        self.assertIsInstance(response.context["form"], EditUserForm)

        form = response.context["form"]
        self.assertEqual(form.initial["first_name"], self.user.first_name)
        self.assertEqual(form.initial["last_name"], self.user.last_name)

    def test_edit_account(self):
        """The user's information should be updated."""
        self.client.force_login(self.user)
        response = self.client.post(
            self.edit_account_url,
            {
                "first_name": "New",
                "last_name": "Name",
            },
            follow=True,
        )

        # User's information should be updated
        self.user.refresh_from_db()
        self.assertEqual(self.user.first_name, "New")
        self.assertEqual(self.user.last_name, "Name")

        # Success message should be displayed
        self.assertRedirects(
            response, reverse("hidp_account_management:edit_account_done")
        )
        self.assertTemplateUsed(
            response, "hidp/accounts/management/edit_account_done.html"
        )


class TestPasswordChangeView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.change_password_url = reverse("hidp_account_management:change_password")

    def test_login_required(self):
        """Anonymous users should be redirected to the login page."""
        self.client.logout()
        response = self.client.get(self.change_password_url)
        self.assertRedirects(
            response,
            f"{reverse('hidp_accounts:login')}?next={self.change_password_url}",
        )

    def test_redirect_user_without_usable_password(self):
        """Users without a usable password should be redirected to the manage page."""
        self.user.set_unusable_password()
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.get(self.change_password_url, follow=True)
        self.assertRedirects(response, reverse("hidp_account_management:set_password"))

    def test_get(self):
        """The password change page should be displayed for authenticated users."""
        self.client.force_login(self.user)
        response = self.client.get(self.change_password_url)
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(
            response, "hidp/accounts/management/password_change.html"
        )
        self.assertIn("form", response.context)
        self.assertTrue(response.context["form"].fields["old_password"].required)

    def test_change_password(self):
        """The user's password should be updated."""
        self.client.force_login(self.user)
        with (
            self.assertTemplateUsed("hidp/accounts/management/email/password_changed_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/password_changed_body.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/password_changed_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                self.change_password_url,
                {
                    "old_password": "P@ssw0rd!",
                    "new_password1": "new_password",
                    "new_password2": "new_password",
                },
                follow=True,
            )

        # User's password should be updated
        self.user.refresh_from_db()
        self.assertTrue(self.user.check_password("new_password"))

        # Redirect to the success page
        self.assertRedirects(
            response, reverse("hidp_account_management:change_password_done")
        )
        self.assertTemplateUsed("hidp/accounts/management/password_change_done.html")

        # Changed password mail should be sent
        self.assertEqual(1, len(mail.outbox))
        message = mail.outbox[0]
        self.assertEqual(message.to, [self.user.email])
        self.assertEqual("Your password has been changed", message.subject)
        self.assertIn(
            reverse("hidp_accounts:password_reset_request"),
            message.body,
        )


class TestSetPasswordView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.user.set_unusable_password()
        cls.user.save()
        cls.set_password_url = reverse("hidp_account_management:set_password")

    def test_login_required(self):
        """Anonymous users should be redirected to the login page."""
        response = self.client.get(self.set_password_url)
        self.assertRedirects(
            response,
            f"{reverse('hidp_accounts:login')}?next={self.set_password_url}",
        )

    def test_redirect_user_with_usable_password(self):
        """Users without a usable password should be redirected to the manage page."""
        self.user.set_password("P@ssw0rd!")
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.get(self.set_password_url, follow=True)
        self.assertRedirects(
            response, reverse("hidp_account_management:change_password")
        )

    def test_get_not_recently_authenticated(self):
        """The must reauthenticate flag should be set to True."""
        self.client.force_login(self.user)
        self.user.last_login = timezone.now() - timedelta(days=1)
        self.user.save()
        response = self.client.get(self.set_password_url)
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(response, "hidp/accounts/management/set_password.html")
        self.assertTrue(
            response.context["must_reauthenticate"],
            msg="Expected must_reauthenticate to be True.",
        )

    def test_post_not_recently_authenticated(self):
        """User is redirected to the same page to reauthenticate."""
        self.client.force_login(self.user)
        self.user.last_login = timezone.now() - timedelta(days=1)
        self.user.save()
        response = self.client.post(
            self.set_password_url,
            {
                "new_password1": "new_password",
                "new_password2": "new_password",
            },
            follow=True,
        )

        # User's password should not be updated
        self.user.refresh_from_db()
        self.assertFalse(
            self.user.check_password("new_password"),
            msg="Expected password to not be set.",
        )

        # Redirect to the same page
        self.assertRedirects(response, self.set_password_url)

    def test_get(self):
        """The set password page should be displayed for authenticated users."""
        self.client.force_login(self.user)
        response = self.client.get(self.set_password_url)
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(response, "hidp/accounts/management/set_password.html")
        self.assertIn("form", response.context)

    def test_set_password(self):
        """The user's password should be set."""
        self.client.force_login(self.user)
        with (
            self.assertTemplateUsed("hidp/accounts/management/email/password_changed_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/password_changed_body.txt"),
            self.assertTemplateUsed("hidp/accounts/management/email/password_changed_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                self.set_password_url,
                {
                    "new_password1": "new_password",
                    "new_password2": "new_password",
                },
                follow=True,
            )

        # User's password should be updated
        self.user.refresh_from_db()
        self.assertTrue(
            self.user.check_password("new_password"), msg="Expected password to be set."
        )

        # Redirect to the success page
        self.assertRedirects(
            response, reverse("hidp_account_management:set_password_done")
        )
        self.assertTemplateUsed("hidp/accounts/management/set_change_done.html")

        # Changed password mail should be sent
        self.assertEqual(1, len(mail.outbox))
        message = mail.outbox[0]
        self.assertEqual(message.to, [self.user.email])
        self.assertEqual("Your password has been changed", message.subject)
        self.assertIn(
            reverse("hidp_accounts:password_reset_request"),
            message.body,
        )



import urllib.parse

from http import HTTPStatus
from unittest import mock

from django.contrib.auth import get_user_model
from django.core import mail
from django.http import HttpRequest
from django.test import TestCase, override_settings
from django.urls import reverse

from hidp.config import configure_oidc_clients
from hidp.federated import models, views
from hidp.federated.constants import OIDC_STATES_SESSION_KEY
from hidp.federated.oidc.exceptions import InvalidOIDCStateError, OAuth2Error, OIDCError
from hidp.test.factories import user_factories

from ...unit_tests.test_federated.test_providers.example import (
    ExampleOIDCClient,
    code_challenge_from_code_verifier,
)

UserModel = get_user_model()


class TestOIDCAuthenticationRequestView(TestCase):
    def setUp(self):
        configure_oidc_clients(ExampleOIDCClient(client_id="test"))

    def test_requires_https(self):
        response = self.client.get(
            reverse(
                "hidp_oidc_client:authenticate", kwargs={"provider_key": "example"}
            ),
            secure=False,
        )
        self.assertEqual(response.status_code, 400)

    def test_requires_post(self):
        response = self.client.get(
            reverse(
                "hidp_oidc_client:authenticate", kwargs={"provider_key": "example"}
            ),
            secure=True,
        )
        self.assertEqual(response.status_code, 405)

    def test_unknown_provider(self):
        response = self.client.post(
            reverse(
                "hidp_oidc_client:authenticate", kwargs={"provider_key": "unknown"}
            ),
            secure=True,
        )
        self.assertEqual(response.status_code, 404)

    def test_redirects_to_provider(self):
        response = self.client.post(
            reverse(
                "hidp_oidc_client:authenticate", kwargs={"provider_key": "example"}
            ),
            secure=True,
        )
        state_key = next(iter(self.client.session[OIDC_STATES_SESSION_KEY]))
        code_verifier = self.client.session[OIDC_STATES_SESSION_KEY][state_key][
            "code_verifier"
        ]
        code_challenge = code_challenge_from_code_verifier(code_verifier)
        callback_url = urllib.parse.quote(
            "https://testserver"
            + reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"})
        )
        self.assertRedirects(
            response,
            (
                f"https://example.com/auth"
                f"?client_id=test"
                f"&response_type=code"
                f"&scope=openid+email+profile"
                f"&redirect_uri={callback_url}"
                f"&state={state_key}"
                f"&code_challenge={code_challenge}"
                f"&code_challenge_method=S256"
            ),
            fetch_redirect_response=False,
        )

    def test_stores_next_url_with_state(self):
        self.client.post(
            reverse(
                "hidp_oidc_client:authenticate", kwargs={"provider_key": "example"}
            ),
            {"next": "/next"},
            secure=True,
        )
        state_key = next(iter(self.client.session[OIDC_STATES_SESSION_KEY]))
        self.assertEqual(
            self.client.session[OIDC_STATES_SESSION_KEY][state_key]["next_url"], "/next"
        )

    def test_reauthenticate_params(self):
        response = self.client.post(
            reverse(
                "hidp_oidc_client:reauthenticate", kwargs={"provider_key": "example"}
            ),
            secure=True,
        )
        state_key = next(iter(self.client.session[OIDC_STATES_SESSION_KEY]))
        code_verifier = self.client.session[OIDC_STATES_SESSION_KEY][state_key][
            "code_verifier"
        ]
        code_challenge = code_challenge_from_code_verifier(code_verifier)
        callback_url = urllib.parse.quote(
            "https://testserver"
            + reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"})
        )
        self.assertRedirects(
            response,
            (
                f"https://example.com/auth"
                f"?client_id=test"
                f"&response_type=code"
                f"&scope=openid+email+profile"
                f"&redirect_uri={callback_url}"
                f"&state={state_key}"
                f"&code_challenge={code_challenge}"
                f"&code_challenge_method=S256"
                # Adds prompt=login and max_age=0 to force reauthentication
                f"&prompt=login"
                f"&max_age=0"
            ),
            fetch_redirect_response=False,
        )


_VALID_AUTH_CALLBACK = (
    {
        "id_token": "id_token",
        "access_token": "access_token",
        "token_type": "token_type",
    },
    {
        "iss": "example",
        "sub": "test_subject",
        "email": "user@example.com",
    },
    {
        "given_name": "Firstname",
        "family_name": "Lastname",
    },
)


@override_settings(REGISTRATION_ENABLED=True)
class TestOIDCAuthenticationCallbackView(TestCase):
    def setUp(self):
        configure_oidc_clients(ExampleOIDCClient(client_id="test"))

    def test_requires_https(self):
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"}),
            secure=False,
        )
        self.assertEqual(response.status_code, 400)

    def test_unknown_provider(self):
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "unknown"}),
            secure=True,
        )
        self.assertEqual(response.status_code, 404)

    @mock.patch(
        "hidp.federated.views.authorization_code_flow.handle_authentication_callback",
        return_value=(*_VALID_AUTH_CALLBACK, None),
    )
    def test_calls_handle_authentication_callback(
        self, mock_handle_authentication_callback
    ):
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"}),
            secure=True,
            follow=False,
        )
        mock_handle_authentication_callback.assert_called_once()
        # Redirects to the next url
        self.assertEqual(response.status_code, HTTPStatus.FOUND)

    @mock.patch(
        "hidp.federated.views.authorization_code_flow.handle_authentication_callback",
        return_value=(*_VALID_AUTH_CALLBACK, "/next"),
    )
    def test_restores_next_url(self, mock_handle_authentication_callback):
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"}),
            secure=True,
            follow=False,
        )
        query = urllib.parse.parse_qs(urllib.parse.urlparse(response.url).query)
        self.assertIn("next", query)
        self.assertEqual(query["next"][0], "/next")

    @mock.patch(
        "hidp.federated.views.authorization_code_flow.handle_authentication_callback",
        side_effect=InvalidOIDCStateError("OIDC state not found"),
    )
    def test_handles_state_error(self, mock_handle_authentication_callback):
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"}),
            secure=True,
            follow=True,
        )
        self.assertTemplateUsed(response, "hidp/accounts/login.html")
        self.assertInHTML(
            "The authentication request has expired. Please try again.",
            response.content.decode("utf-8"),
        )

    @mock.patch(
        "hidp.federated.views.authorization_code_flow.handle_authentication_callback",
        side_effect=OIDCError("OIDC error"),
    )
    def test_handles_oidc_error(self, mock_handle_authentication_callback):
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"}),
            secure=True,
            follow=True,
        )
        self.assertTemplateUsed(response, "hidp/accounts/login.html")
        self.assertInHTML(
            "An unexpected error occurred during authentication. Please try again.",
            response.content.decode("utf-8"),
        )

    @mock.patch(
        "hidp.federated.views.authorization_code_flow.handle_authentication_callback",
        side_effect=OAuth2Error("OAuth2 error"),
    )
    def test_handles_oauth2_error(self, mock_handle_authentication_callback):
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"}),
            secure=True,
            follow=True,
        )
        self.assertTemplateUsed(response, "hidp/accounts/login.html")
        self.assertInHTML(
            "An unexpected error occurred during authentication. Please try again.",
            response.content.decode("utf-8"),
        )

    @mock.patch(
        "hidp.federated.views.authorization_code_flow.handle_authentication_callback",
        return_value=(*_VALID_AUTH_CALLBACK, None),
    )
    def test_redirect_to_register(self, mock_handle_authentication_callback):
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"}),
            secure=True,
        )
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        redirect = urllib.parse.urlparse(response.url)
        self.assertEqual(redirect.path, reverse("hidp_oidc_client:register"))
        query = urllib.parse.parse_qs(redirect.query)
        self.assertIn("token", query)
        token = query["token"][0]
        self.assertIn(token, self.client.session)

    @mock.patch(
        "hidp.federated.views.authorization_code_flow.handle_authentication_callback",
        return_value=(*_VALID_AUTH_CALLBACK, None),
    )
    def test_redirect_to_login(self, mock_handle_authentication_callback):
        user = user_factories.VerifiedUserFactory()
        models.OpenIdConnection.objects.create(
            user=user,
            provider_key="example",
            issuer_claim="example",
            subject_claim="test_subject",
        )
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"}),
            secure=True,
        )
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        redirect = urllib.parse.urlparse(response.url)
        self.assertEqual(redirect.path, reverse("hidp_oidc_client:login"))
        query = urllib.parse.parse_qs(redirect.query)
        self.assertIn("token", query)
        token = query["token"][0]
        self.assertIn(token, self.client.session)

    @mock.patch(
        "hidp.federated.views.authorization_code_flow.handle_authentication_callback",
        return_value=(*_VALID_AUTH_CALLBACK, None),
    )
    def test_must_login_to_link_account(self, mock_handle_authentication_callback):
        # A user with the same email address exists, but is not logged in
        user_factories.UserFactory(email="user@example.com")
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"}),
            secure=True,
            follow=True,
        )
        self.assertInHTML(
            "You already have an account with this email address."
            " Please sign in to link your account.",
            response.content.decode("utf-8"),
        )

    @mock.patch(
        "hidp.federated.views.authorization_code_flow.handle_authentication_callback",
        return_value=(*_VALID_AUTH_CALLBACK, None),
    )
    def test_redirect_to_link_account(self, mock_handle_authentication_callback):
        # A user is logged in, but no connection exists. Continue to link account.
        user = user_factories.VerifiedUserFactory()
        self.client.force_login(user)
        response = self.client.get(
            reverse("hidp_oidc_client:callback", kwargs={"provider_key": "example"}),
            secure=True,
        )
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        redirect = urllib.parse.urlparse(response.url)
        self.assertEqual(redirect.path, reverse("hidp_oidc_management:link_account"))
        query = urllib.parse.parse_qs(redirect.query)
        self.assertIn("token", query)
        token = query["token"][0]
        self.assertIn(token, self.client.session)


class OIDCTokenDataTestMixin:
    view_name = NotImplemented
    view_class = NotImplemented

    @classmethod
    def setUpTestData(cls):
        cls.url = reverse(cls.view_name)

    def setUp(self):
        configure_oidc_clients(ExampleOIDCClient(client_id="test"))

    def _assert_invalid_token(self, *, token=None, method="get"):
        request_method = getattr(self.client, method)
        response = (
            request_method(self.url, follow=True)
            if token is None
            else request_method(self.url, {"token": token}, follow=True)
        )
        self.assertInHTML(
            "Expired or invalid token. Please try again.",
            response.content.decode("utf-8"),
        )

    def _add_oidc_data_to_session(self, *, save=True):
        session = self.client.session
        request = HttpRequest()
        request.session = session
        token = self.view_class.add_data_to_session(
            request,
            provider_key="example",
            claims={
                "iss": "example",
                "sub": "test-subject",
                "email": "user@example.com",
            },
            user_info={
                "given_name": "Firstname",
                "family_name": "Lastname",
            },
        )
        if save:
            session.save()
        return token

    def test_requires_token(self):
        self._assert_invalid_token()

    def test_invalid_token(self):
        self._assert_invalid_token(token="invalid")

    def test_post_invalid_token(self):
        self._assert_invalid_token(token="invalid", method="post")

    def test_valid_token_missing_session_data(self):
        # Do not save the session to mimic an expired session or hijacked token
        token = self._add_oidc_data_to_session(save=False)
        self._assert_invalid_token(token=token)


@override_settings(REGISTRATION_ENABLED=True)
class TestOIDCRegistrationView(OIDCTokenDataTestMixin, TestCase):
    view_class = views.OIDCRegistrationView
    view_name = "hidp_oidc_client:register"

    def test_get_with_valid_token(self):
        token = self._add_oidc_data_to_session()
        response = self.client.get(self.url, {"token": token})
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "hidp/federated/registration.html")

    def test_post_with_valid_token(self):
        token = self._add_oidc_data_to_session()
        with (
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_body.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                self.url + f"?token={token}",
                {
                    "first_name": "Firstname",
                    "last_name": "Lastname",
                    "agreed_to_tos": "on",
                },
                follow=True,
            )
        user = UserModel.objects.filter(email="user@example.com").first()
        self.assertIsNotNone(user, msg="Expected a user to be created.")

        self.assertIsNone(user.email_verified, msg="Expected email to be unverified.")

        # Verification email sent
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(
            message.subject,
            "Verify your email address",
        )
        # Redirected to verification required page
        self.assertRedirects(
            response,
            reverse(
                "hidp_accounts:email_verification_required", kwargs={"token": "email"}
            ),
        )
        # Verification required page
        self.assertInHTML(
            "Verification required",
            response.content.decode("utf-8"),
        )


class TestOIDCLoginView(OIDCTokenDataTestMixin, TestCase):
    view_class = views.OIDCLoginView
    view_name = "hidp_oidc_client:login"

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = user_factories.VerifiedUserFactory()
        cls.connection = models.OpenIdConnection.objects.create(
            user=cls.user,
            provider_key="example",
            issuer_claim="example",
            subject_claim="test-subject",
        )

    def test_valid_login(self):
        token = self._add_oidc_data_to_session()
        response = self.client.get(self.url, {"token": token})
        self.assertEqual(response.wsgi_request.user, self.user)

    def test_valid_login_inactive_user(self):
        self.user.is_active = False
        self.user.save()
        token = self._add_oidc_data_to_session()
        response = self.client.get(self.url, {"token": token}, follow=True)
        self.assertTemplateUsed(response, "hidp/accounts/login.html")
        self.assertInHTML(
            "Login failed. Invalid credentials.",
            response.content.decode("utf-8"),
        )

    def test_valid_login_unverified_user(self):
        self.user.email_verified = None
        self.user.save()
        token = self._add_oidc_data_to_session()
        with (
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_body.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_body.html"),
        ):  # fmt: skip
            response = self.client.get(self.url, {"token": token}, follow=True)

        # Verification email sent
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(
            message.subject,
            "Verify your email address",
        )
        # Redirected to verification required page
        self.assertRedirects(
            response,
            reverse(
                "hidp_accounts:email_verification_required", kwargs={"token": "email"}
            ),
        )
        # Verification required page
        self.assertInHTML(
            "Verification required",
            response.content.decode("utf-8"),
        )


class TestOIDCClient(ExampleOIDCClient):
    # Same as the ExampleOIDCClient, but with a different provider key.
    provider_key = "test"


class TestOIDCLinkedServicesView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.user.set_unusable_password()
        cls.user.save()
        cls.oidc_linked_services_url = reverse("hidp_oidc_management:linked_services")

    def setUp(self):
        clients = [
            ExampleOIDCClient(client_id="example"),
            TestOIDCClient(client_id="test"),
        ]
        configure_oidc_clients(*clients)

    def test_login_required(self):
        """Anonymous users should be redirected to the login page."""
        response = self.client.get(self.oidc_linked_services_url)
        self.assertRedirects(
            response,
            f"{reverse('hidp_accounts:login')}?next={self.oidc_linked_services_url}",
        )

    def test_available_services(self):
        self.client.force_login(self.user)
        response = self.client.get(self.oidc_linked_services_url)

        # No services linked
        self.assertInHTML("Linked services", response.content.decode("utf-8"), count=0)

        # List of available services should be displayed
        self.assertInHTML(
            "Available services",
            response.content.decode("utf-8"),
        )
        for name in ("Example", "Test"):
            self.assertInHTML(
                f"<button type='submit'>Link with {name}</button>",
                response.content.decode("utf-8"),
            )

    def test_linked_one_service(self):
        models.OpenIdConnection.objects.create(
            user=self.user,
            provider_key="example",
            issuer_claim="example",
            subject_claim="test-subject",
        )

        self.client.force_login(self.user)
        response = self.client.get(self.oidc_linked_services_url)

        # Unlinking is not possible with only one service linked and no password set
        self.assertFalse(
            response.context["can_unlink"], msg="Expected can_unlink to be False."
        )

        # The linked service should be displayed
        self.assertInHTML(
            "Linked services",
            response.content.decode("utf-8"),
        )

        # Unlink option should be disabled
        self.assertInHTML(
            "<button type='submit' disabled>Unlink from Example</button>",
            response.content.decode("utf-8"),
        )

        # The remaining service should be available
        self.assertInHTML(
            "Available services",
            response.content.decode("utf-8"),
        )
        self.assertInHTML(
            "<button type='submit'>Link with Test</button>",
            response.content.decode("utf-8"),
        )

    def test_linked_all_services(self):
        models.OpenIdConnection.objects.create(
            user=self.user,
            provider_key="example",
            issuer_claim="example",
            subject_claim="test-subject",
        )
        models.OpenIdConnection.objects.create(
            user=self.user,
            provider_key="test",
            issuer_claim="example",
            subject_claim="test-subject",
        )

        self.client.force_login(self.user)
        response = self.client.get(self.oidc_linked_services_url)

        # Unlinking is possible with multiple services linked
        self.assertTrue(
            response.context["can_unlink"], msg="Expected can_unlink to be True."
        )

        self.assertInHTML(
            "Linked services",
            response.content.decode("utf-8"),
        )
        for key in ("example", "test"):
            self.assertInHTML(
                f"<button type='submit'>Unlink from {key.capitalize()}</button>",
                response.content.decode("utf-8"),
            )

        # No additional services should be available
        self.assertInHTML(
            "Available services", response.content.decode("utf-8"), count=0
        )

    def test_linked_one_service_with_password_set(self):
        self.user.set_password("P@ssw0rd!")
        self.user.save()

        models.OpenIdConnection.objects.create(
            user=self.user,
            provider_key="example",
            issuer_claim="example",
            subject_claim="test-subject",
        )

        self.client.force_login(self.user)

        response = self.client.get(self.oidc_linked_services_url)

        # Unlinking is possible with one service linked and a password set
        self.assertTrue(
            response.context["can_unlink"], msg="Expected can_unlink to be True."
        )

        self.assertInHTML(
            "Linked services",
            response.content.decode("utf-8"),
        )
        self.assertInHTML(
            "<button type='submit'>Unlink from Example</button>",
            response.content.decode("utf-8"),
        )


class TestOIDCAccountLinkView(OIDCTokenDataTestMixin, TestCase):
    view_class = views.OIDCAccountLinkView
    view_name = "hidp_oidc_management:link_account"

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = user_factories.VerifiedUserFactory()

    def setUp(self):
        super().setUp()
        self.client.force_login(self.user)

    def test_requires_login(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertRedirects(
            response,
            f"{reverse('hidp_accounts:login')}?next={self.url}",
        )

    def test_get_with_valid_token(self):
        token = self._add_oidc_data_to_session()
        response = self.client.get(self.url, {"token": token})
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "hidp/federated/account_link.html")

    def test_post_with_valid_token(self):
        token = self._add_oidc_data_to_session()
        response = self.client.post(
            self.url + f"?token={token}",
            {"allow_link": "on"},
            follow=True,
        )
        connection = models.OpenIdConnection.objects.filter(user=self.user).first()
        self.assertIsNotNone(connection, msg="Expected connection to be created.")

        # Redirected to the done page
        self.assertRedirects(
            response,
            reverse(
                "hidp_oidc_management:link_account_done",
                kwargs={"provider_key": "example"},
            ),
        )
        self.assertTemplateUsed(response, "hidp/federated/account_link_done.html")


class TestOIDCAccountUnlinkView(TestCase):
    view_name = "hidp_oidc_management:unlink_account"

    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()
        cls.url = reverse(cls.view_name, kwargs={"provider_key": "example"})
        configure_oidc_clients(ExampleOIDCClient(client_id="test"))
        cls.connection = models.OpenIdConnection.objects.create(
            user=cls.user,
            provider_key="example",
            issuer_claim="example",
            subject_claim="test-subject",
        )

    def setUp(self):
        super().setUp()
        self.client.force_login(self.user)

    def test_requires_login(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertRedirects(
            response,
            f"{reverse('hidp_accounts:login')}?next={self.url}",
        )

    def test_get_with_invalid_provider(self):
        response = self.client.get(
            reverse(self.view_name, kwargs={"provider_key": "unknown"})
        )
        self.assertEqual(response.status_code, 404)

    def test_get_with_valid_provider(self):
        response = self.client.get(self.url)
        self.assertTemplateUsed(response, "hidp/federated/account_unlink.html")
        self.assertEqual(response.status_code, 200)

    def test_post_with_valid_provider(self):
        response = self.client.post(self.url, {"allow_unlink": "on"}, follow=True)

        # Redirected to the done page
        self.assertRedirects(
            response,
            reverse(
                "hidp_oidc_management:unlink_account_done",
                kwargs={"provider_key": "example"},
            ),
        )
        self.assertTemplateUsed(response, "hidp/federated/account_unlink_done.html")

    def test_delete_only_login_method(self):
        # Remove the user's password so they can only log in via OIDC
        self.user.set_unusable_password()
        self.user.save()
        self.client.force_login(self.user)
        # Attempt to unlink the only login method
        response = self.client.post(self.url, {"allow_unlink": "on"}, follow=True)
        # Form error
        self.assertTemplateUsed(response, "hidp/federated/account_unlink.html")
        self.assertFormError(
            response.context["form"],
            None,
            "You cannot unlink your only way to sign in.",
        )

    def test_valid_provider_no_connection(self):
        other_user = user_factories.VerifiedUserFactory()
        self.client.force_login(other_user)
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 404)

        response = self.client.post(self.url, {"allow_unlink": "on"})
        self.assertEqual(response.status_code, 404)



import json

from datetime import UTC, datetime, timedelta
from hashlib import sha256
from http import HTTPStatus
from urllib.parse import parse_qs, urlencode, urlsplit

from jwcrypto import jwk, jwt
from oauth2_provider.models import get_access_token_model, get_application_model

from django.conf import settings
from django.core.signing import b64_encode
from django.test import TestCase, override_settings
from django.urls import reverse
from django.utils.timezone import now as tz_now

from hidp.test.factories import user_factories

Application = get_application_model()
AccessToken = get_access_token_model()


@override_settings(REGISTRATION_ENABLED=True)
class TestOAuthFlow(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory(
            first_name="Henk", last_name="de Vries", email="henk@example.com"
        )
        # Create a public application (client), that does not keep a secret.
        #
        # * Use the Authorization Code grant type (PKCE is enabled by default).
        # * Skip authorization, this client is trusted (e.g. a first-party app).
        # * Use the RS256 algorithm for ID tokens. This allows the client to
        #   verify the token signature using the public key.
        #
        # This is the preferred setup for any (first-party) client application.
        cls.trusted_application = Application.objects.create(
            name="Happy App",
            client_id="happy-app",
            client_type=Application.CLIENT_PUBLIC,
            client_secret="",
            authorization_grant_type=Application.GRANT_AUTHORIZATION_CODE,
            skip_authorization=True,
            redirect_uris="https://127.0.0.1/",
            algorithm=Application.RS256_ALGORITHM,
        )
        cls.third_party_application = Application.objects.create(
            name="Shady App",
            client_id="shady-app",
            client_type=Application.CLIENT_PUBLIC,
            client_secret="",
            authorization_grant_type=Application.GRANT_AUTHORIZATION_CODE,
            skip_authorization=False,
            redirect_uris="https://127.0.0.1/",
            algorithm=Application.RS256_ALGORITHM,
        )

    def set_client_access_token(self, client=None, user=None, scope=""):
        """Add an access token to the test client."""
        # Utility method to add an access token to the test client, used in
        # the test methods to simulate a logged-in user.

        token = AccessToken.objects.create(
            user=user or self.user,
            scope=scope,
            expires=tz_now() + timedelta(seconds=300),
            token="secret-access-token-key",
            application=self.trusted_application,
        )
        (client or self.client).defaults["HTTP_AUTHORIZATION"] = f"Bearer {token}"

    def authorization_request(
        self, code_verifier="secret", client_id="happy-app", **oauth_params
    ):
        """Perform an authorization request."""
        # Utility method to perform an authorization request, used in the test
        # methods to simulate the authorization code grant flow.

        # PKCE challenge
        if code_verifier:
            code_challenge = b64_encode(sha256(code_verifier.encode()).digest())
        else:
            code_challenge = None

        # Authorization request: code grant + PKCE
        return self.client.get(
            "/o/authorize/",
            # Filter out None values
            {
                key: value
                for key, value in {
                    "response_type": "code",
                    "client_id": client_id,
                    "scope": "openid profile email",
                    "redirect_uri": "https://127.0.0.1/",
                    "code_challenge": code_challenge,
                    "code_challenge_method": "S256",
                    **oauth_params,
                }.items()
                if value is not None
            },
        )

    def test_authorize_authentication_required(self):
        """Authorization endpoint requires authentication."""
        # User is not logged in
        response = self.authorization_request()

        self.assertEqual(HTTPStatus.FOUND, response.status_code)
        self.assertRegex(
            response["Location"],
            r"^/login/\?next=/o/authorize/"
            r"%3Fresponse_type%3Dcode"
            r"%26client_id%3Dhappy-app"
            r"%26scope%3Dopenid%2Bprofile%2Bemail"
            r"%26redirect_uri%3Dhttps%253A%252F%252F127.0.0.1%252F"
            r"%26code_challenge%3D[a-zA-Z0-9-_]+"
            r"%26code_challenge_method%3DS256$",
        )

    def test_requires_pkce(self):
        self.client.force_login(self.user)  # User is logged in
        response = self.authorization_request(None)
        self.assertRedirects(
            response,
            "https://127.0.0.1/?error=invalid_request&error_description=Code+challenge+required.",
            fetch_redirect_response=False,
        )

    def test_code_grant_token_exchange(self):
        """Test authorization flow."""
        code_verifier = "secret"
        self.client.force_login(self.user)  # User is logged in
        response = self.authorization_request(code_verifier)

        self.assertEqual(HTTPStatus.FOUND, response.status_code)
        self.assertRegex(
            response["Location"], r"^https://127\.0\.0\.1/\?code=[A-z0-9]+$"
        )

        # Get authorization code
        code = parse_qs(urlsplit(response["Location"]).query)["code"]

        # Trade authorization code for tokens
        response = self.client.post(
            "/o/token/",
            {
                "client_id": "happy-app",
                "grant_type": "authorization_code",
                "code": code,
                "code_verifier": code_verifier,
            },
        )
        self.assertEqual(HTTPStatus.OK, response.status_code)
        self.assertEqual(
            {
                "access_token",
                "refresh_token",
                "id_token",
                "expires_in",
                "token_type",
                "scope",
            },
            set(response.json().keys()),
        )

        # Inspect id_token
        key = jwk.JWK.from_pem(
            settings.OAUTH2_PROVIDER["OIDC_RSA_PRIVATE_KEY"].encode()
        )
        id_token = jwt.JWT(jwt=response.json()["id_token"], key=key)
        claims = json.loads(id_token.claims)
        self.assertEqual(
            {
                "aud",
                "iat",
                "at_hash",
                "sub",
                "iss",
                "exp",
                "auth_time",
                "jti",
                "email_verified",
                "email",
                "family_name",
                "updated_at",
                "name",
                "given_name",
            },
            set(claims.keys()),
        )
        self.assertEqual("happy-app", claims["aud"])
        self.assertEqual(str(self.user.id), claims["sub"])

    def test_authorize_login_prompt(self):
        """Test authorization endpoint can force login prompt."""
        self.client.force_login(self.user)  # User is logged in
        response = self.authorization_request(prompt="login")

        self.assertEqual(HTTPStatus.FOUND, response.status_code)
        self.assertRegex(
            response["Location"],
            r"^/login/\?next=/o/authorize/"
            r"%3Fresponse_type%3Dcode"
            r"%26client_id%3Dhappy-app"
            r"%26scope%3Dopenid%2Bprofile%2Bemail"
            r"%26redirect_uri%3Dhttps%253A%252F%252F127.0.0.1%252F"
            r"%26code_challenge%3D[a-zA-Z0-9-_]+"
            r"%26code_challenge_method%3DS256$",
        )

    def test_authorize_prompt_none(self):
        """Test authorization endpoint can skip login prompt."""
        with self.subTest("User is not logged in"):
            response = self.authorization_request(prompt="none")
            self.assertEqual(HTTPStatus.FOUND, response.status_code)
            self.assertURLEqual(
                response["Location"], "https://127.0.0.1/?error=login_required"
            )

        with self.subTest("User is logged in"):
            self.client.force_login(self.user)
            response = self.authorization_request(prompt="none")
            self.assertEqual(HTTPStatus.FOUND, response.status_code)
            self.assertRegex(
                response["Location"], r"^https://127\.0\.0\.1/\?code=[A-z0-9]+$"
            )

        with self.subTest("Consent required"):
            # User is logged in, but consent is required
            response = self.authorization_request(prompt="none", client_id="shady-app")
            self.assertEqual(HTTPStatus.FOUND, response.status_code)
            self.assertURLEqual(
                response["Location"], "https://127.0.0.1/?error=consent_required"
            )

    def test_authorize_create_prompt(self):
        """prompt=create redirects to registration."""
        with self.subTest("No user logged in"):
            response = self.authorization_request(prompt="create")
            next_url = (
                f"{response.request['PATH_INFO']}?{response.request['QUERY_STRING']}"
            ).replace("&prompt=create", "")
            self.assertRedirects(
                response,
                (
                    f"{reverse('hidp_accounts:register')}"
                    f"?{urlencode({'next': next_url})}"
                ),
            )

        with self.subTest("User logged in"):
            self.client.force_login(self.user)
            response = self.authorization_request(prompt="create")
            next_url = (
                f"{response.request['PATH_INFO']}?{response.request['QUERY_STRING']}"
            ).replace("&prompt=create", "")
            self.assertRedirects(
                response,
                (
                    f"{reverse('hidp_accounts:register')}"
                    f"?{urlencode({'next': next_url})}"
                ),
            )

    def test_userinfo_limited_scope(self):
        self.set_client_access_token(scope="openid")
        response = self.client.get("/o/userinfo/")
        self.assertEqual(HTTPStatus.OK, response.status_code)
        userinfo = response.json()
        self.assertEqual({"sub"}, set(userinfo.keys()))
        self.assertEqual(str(self.user.id), userinfo["sub"])

    def test_userinfo_profile_scope(self):
        self.set_client_access_token(scope="openid profile")
        response = self.client.get("/o/userinfo/")
        self.assertEqual(HTTPStatus.OK, response.status_code)
        userinfo = response.json()
        self.assertEqual(
            {"sub", "name", "given_name", "family_name", "updated_at"},
            set(userinfo.keys()),
        )
        self.assertEqual("Henk de Vries", userinfo["name"])
        self.assertEqual("Henk", userinfo["given_name"])
        self.assertEqual("de Vries", userinfo["family_name"])
        # updated_at is unix time in seconds, so discard the microseconds bit
        self.assertEqual(
            self.user.last_modified.replace(microsecond=0),
            datetime.fromtimestamp(userinfo["updated_at"], tz=UTC),
        )

    def test_userinfo_email_scope(self):
        self.set_client_access_token(scope="openid email")
        response = self.client.get("/o/userinfo/")
        self.assertEqual(HTTPStatus.OK, response.status_code)
        userinfo = response.json()
        self.assertEqual({"sub", "email", "email_verified"}, set(userinfo.keys()))
        self.assertEqual(self.user.email, userinfo["email"])
        self.assertFalse(
            userinfo["email_verified"], msg="Expected email_verified to be False"
        )

    def test_userinfo_all_scopes(self):
        # Also test that setting email_verified results in the correct output
        self.user.email_verified = tz_now()
        self.user.save()
        self.set_client_access_token(scope="openid profile email")
        response = self.client.get("/o/userinfo/")
        self.assertEqual(HTTPStatus.OK, response.status_code)
        userinfo = response.json()
        self.assertEqual(
            {
                "sub",
                "name",
                "given_name",
                "family_name",
                "updated_at",
                "email",
                "email_verified",
            },
            set(userinfo.keys()),
        )
        self.assertTrue(
            userinfo["email_verified"], msg="Expected email_verified to be True"
        )



from datetime import timedelta
from enum import StrEnum, auto

from django.db.models import Q
from django.utils import timezone

from .models import EmailChangeRequest


class Recipient(StrEnum):
    CURRENT_EMAIL = auto()
    PROPOSED_EMAIL = auto()


def remove_complete_and_stale_email_change_requests(*, days=7, dry_run=False):
    """
    Remove completed and stale email change requests.

    Args:
        days (int):
            Maximum number of days an email change request is allowed to be
            incomplete. Defaults to 7.

        dry_run (bool):
            If `True`, returns the number of email change requests that would be
            removed, without performing the removal. Defaults to `False`.

    Returns:
        int:
            The number of email change requests that are deleted, or would be deleted
            if `dry_run` is `True`
    """
    complete_and_stale_email_change_requests = EmailChangeRequest.objects.filter(
        Q(confirmed_by_current_email=True, confirmed_by_proposed_email=True)
        | Q(created_at__lt=timezone.now() - timedelta(days=days)),
    )
    complete_and_stale_email_change_requests_count = (
        complete_and_stale_email_change_requests.count()
    )

    if not dry_run:
        complete_and_stale_email_change_requests.delete()

    return complete_and_stale_email_change_requests_count



from django.template import RequestContext
from django.test import RequestFactory, TestCase

from hidp.csp.templatetags.csp_nonce import csp_nonce


class TestCSPNonceTemplatetag(TestCase):
    client_class = RequestFactory

    def test_hidp_csp_nonce(self):
        """The templatetag returns the nonce."""
        request = self.client.get("/")
        request.hidp_csp_nonce = "nonce"
        context = RequestContext(request)

        self.assertEqual(csp_nonce(context), "nonce")

    def test_no_nonce(self):
        """No nonce."""
        request = self.client.get("/")
        context = RequestContext(request)
        self.assertIsNone(csp_nonce(context))



class OAuth2Error(Exception):
    def __init__(self, error, description=None, uri=None):
        self.error = error
        self.description = description
        self.uri = uri

    def __str__(self):
        message = self.error
        if self.description:
            message += f": {self.description}"
        if self.uri:
            message += f" ({self.uri})"
        return message


class OIDCError(OAuth2Error):
    pass


class InvalidOIDCStateError(OIDCError):
    pass



"""
Authentication URLs.

Provides the URL patterns for the base accounts features.

Include this module in the root URL configuration:

    from hidp.accounts import account_urls

    urlpatterns = [
        path("", include(account_urls)),
    ]

This module also defines the namespace `hidp_accounts` for these URLs.

Include this namespace when reversing URLs, for example:

    reverse("hidp_accounts:login")
"""

from django.urls import include, path

from . import views

app_name = "hidp_accounts"

register_urls = [
    path("signup/", views.RegistrationView.as_view(), name="register"),
    path("terms-of-service/", views.TermsOfServiceView.as_view(), name="tos"),
]

verifications_urls = [
    path(
        "verify/<token>/sent/",
        views.EmailVerificationRequiredView.as_view(),
        name="email_verification_required",
    ),
    path(
        "verify/<token>/verify/",
        views.EmailVerificationView.as_view(),
        name="verify_email",
    ),
    path(
        "verify/complete/",
        views.EmailVerificationCompleteView.as_view(),
        name="email_verification_complete",
    ),
]

auth_urls = [
    path("login/", views.LoginView.as_view(), name="login"),
    path("logout/", views.LogoutView.as_view(), name="logout"),
]

recover_password_urls = [
    path(
        "",
        views.PasswordResetRequestView.as_view(),
        name="password_reset_request",
    ),
    path(
        "sent/",
        views.PasswordResetEmailSentView.as_view(),
        name="password_reset_email_sent",
    ),
    path(
        "<uidb64>/<token>/",
        views.PasswordResetView.as_view(),
        name="password_reset",
    ),
    path(
        "complete/",
        views.PasswordResetCompleteView.as_view(),
        name="password_reset_complete",
    ),
]

recover_urls = [
    path(
        "recover/password/",
        include(recover_password_urls),
    )
]

urlpatterns = register_urls + verifications_urls + auth_urls + recover_urls



from django.core.management import BaseCommand

from ...email_change import remove_complete_and_stale_email_change_requests


class Command(BaseCommand):
    DEFAULT_MAX_DAYS = 7
    help = (
        "Remove completed email change requests and requests that have not been"
        " completed within a specific number of days after creation"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--days",
            type=int,
            default=self.DEFAULT_MAX_DAYS,
            help=(
                f"Maximum number of days an email change request is allowed to be"
                f" incomplete: defaults to {self.DEFAULT_MAX_DAYS}."
            ),
        )

        parser.add_argument(
            "--dry-run",
            action="store_true",
            help=(
                "Show the number of email change requests that would be removed,"
                " without performing the removal."
            ),
        )

    def handle(self, *args, days, dry_run, **options):
        self.stdout.write(
            f"Removing completed email change requests and requests that have not been"
            f" completed within {days} days..."
        )

        complete_and_stale_email_change_requests_count = (
            remove_complete_and_stale_email_change_requests(days=days, dry_run=dry_run)
        )

        if dry_run:
            self.stdout.write(
                self.style.WARNING(
                    f"{complete_and_stale_email_change_requests_count} completed and/or"
                    " stale email change request(s) would be removed."
                )
            )
        else:
            self.stdout.write(
                self.style.SUCCESS(
                    f"Successfully removed"
                    f" {complete_and_stale_email_change_requests_count} completed"
                    f" and/or stale email change request(s)."
                )
            )



from datetime import timedelta

from oauth2_provider.models import get_access_token_model, get_application_model
from rest_framework.test import APITestCase

from django.urls import reverse
from django.utils.timezone import now as tz_now

from hidp.test.factories import user_factories

AccessToken = get_access_token_model()
Application = get_application_model()


class TestUserViewSetViaSession(APITestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory(first_name="Walter", last_name="White")
        cls.url = reverse("api:user-detail", kwargs={"pk": "me"})

    def setUp(self):
        self.client.force_login(self.user)

    def test_get_unauthenticated(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 403)

    def test_get_other_user_not_allowed(self):
        other_user = user_factories.UserFactory()
        self.client.force_login(other_user)

        response = self.client.get(
            reverse("api:user-detail", kwargs={"pk": self.user.pk}),
        )
        self.assertEqual(response.status_code, 404)

    def test_get(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            {
                "first_name": "Walter",
                "last_name": "White",
            },
            response.json(),
        )

    def test_update_user_unauthenticated(self):
        self.client.logout()
        response = self.client.patch(
            self.url,
            data={"first_name": "Skyler"},
        )
        self.assertEqual(response.status_code, 403)

    def test_update_with_pk_not_allowed(self):
        response = self.client.patch(
            reverse("api:user-detail", kwargs={"pk": self.user.pk}),
            data={"first_name": "Skyler"},
        )
        self.assertEqual(response.status_code, 404)

    def test_update_other_user_not_allowed(self):
        other_user = user_factories.UserFactory()
        self.client.force_login(other_user)

        response = self.client.patch(
            reverse("api:user-detail", kwargs={"pk": self.user.pk}),
            data={"first_name": "Skyler"},
        )
        self.assertEqual(response.status_code, 404)

    def test_update_user_with_patch_without_all_required_fields(self):
        # Patch without all required fields should partially update.
        response = self.client.patch(
            self.url,
            data={"first_name": "Skyler"},
        )
        self.assertEqual(response.status_code, 200)
        self.user.refresh_from_db()
        self.assertEqual(self.user.first_name, "Skyler")
        self.assertEqual(self.user.last_name, "White")

    def test_update_user_with_patch_with_all_required_fields(self):
        # Patch with all required fields should update the user."
        response = self.client.patch(
            self.url,
            data={"first_name": "Jesse", "last_name": "Pinkman"},
        )
        self.assertEqual(response.status_code, 200)
        self.user.refresh_from_db()
        self.assertEqual(self.user.first_name, "Jesse")
        self.assertEqual(self.user.last_name, "Pinkman")

    def test_update_user_with_put_without_all_required_fields(self):
        # Put without all required fields should throw an error.
        response = self.client.put(
            self.url,
            data={"first_name": "Skyler"},
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            '{"last_name":["This field is required."]}',
            response.content.decode("utf-8"),
        )
        self.user.refresh_from_db()
        self.assertEqual(self.user.first_name, "Walter")

    def test_update_user_with_put_with_all_required_fields(self):
        # Put with all required fields should update the user.
        response = self.client.put(
            self.url,
            data={"first_name": "Jesse", "last_name": "Pinkman"},
        )
        self.assertEqual(response.status_code, 200)
        self.user.refresh_from_db()
        self.assertEqual(self.user.first_name, "Jesse")
        self.assertEqual(self.user.last_name, "Pinkman")


class TestUserViewSetViaAccessToken(APITestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory(first_name="Walter", last_name="White")
        cls.url = reverse("api:user-detail", kwargs={"pk": "me"})
        cls.trusted_application = Application.objects.create(
            name="Happy App",
            client_id="happy-app",
            client_type=Application.CLIENT_PUBLIC,
            client_secret="",
            authorization_grant_type=Application.GRANT_AUTHORIZATION_CODE,
            skip_authorization=True,
            redirect_uris="https://127.0.0.1/",
            algorithm=Application.RS256_ALGORITHM,
        )

    def set_client_access_token(self, expires_in=300):
        """Add an access token to the test client."""
        # Utility method to add an access token to the test client, used in
        # the test methods to simulate a logged-in user.
        token = AccessToken.objects.create(
            user=self.user,
            scope="openid profile email",
            expires=tz_now() + timedelta(seconds=expires_in),
            token="secret-access-token-key",
            application=self.trusted_application,
        )
        self.client.defaults["HTTP_AUTHORIZATION"] = f"Bearer {token}"

    def test_get_with_expired_token(self):
        self.set_client_access_token(expires_in=-300)

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 403)

    def test_get_with_access_token(self):
        self.set_client_access_token()

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            {
                "first_name": "Walter",
                "last_name": "White",
            },
            response.json(),
        )

    def test_update_user_with_expired_token(self):
        self.set_client_access_token(expires_in=-300)

        response = self.client.patch(
            self.url,
            data={"first_name": "Skyler"},
        )
        self.assertEqual(response.status_code, 403)
        self.user.refresh_from_db()
        self.assertEqual(self.user.first_name, "Walter")
        self.assertEqual(self.user.last_name, "White")

    def test_update_user_with_access_token(self):
        self.set_client_access_token()

        response = self.client.patch(
            self.url,
            data={"first_name": "Skyler"},
        )
        self.assertEqual(response.status_code, 200)
        self.user.refresh_from_db()
        self.assertEqual(self.user.first_name, "Skyler")
        self.assertEqual(self.user.last_name, "White")



from django.http import HttpResponse
from django.test import RequestFactory, TestCase

from hidp.csp.decorators import hidp_csp_protection


def regular_view(request):
    return HttpResponse("Regular view")


@hidp_csp_protection
def csp_protected_view(request):
    return HttpResponse("Protected view")


@hidp_csp_protection
def csp_protected_view_with_existing_csp(request):
    return HttpResponse(
        "Protected view with existing CSP",
        headers={"Content-Security-Policy": "existing"},
    )


class TestCSPProtectionDecorator(TestCase):
    client_class = RequestFactory

    def test_regular_view(self):
        request = self.client.get("/")
        response = regular_view(request)
        self.assertFalse(
            "Content-Security-Policy" in response.headers,
            msg="CSP header should not be present for regular views.",
        )

    def test_hidp_csp_protection_view(self):
        request = self.client.get("/")
        response = csp_protected_view(request)
        self.assertTrue(
            "Content-Security-Policy" in response.headers,
            msg="CSP header should be present for decorated views.",
        )

    def test_does_not_override_existing_csp_header(self):
        request = self.client.get("/")
        response = csp_protected_view_with_existing_csp(request)
        self.assertEqual(
            response.headers["Content-Security-Policy"],
            "existing",
            msg="CSP header should not be overridden.",
        )



import io

from unittest import mock

import requests

from jwcrypto.jwk import JWK, JWKSet

from django.core.cache import cache
from django.core.management import call_command
from django.test import SimpleTestCase, TestCase, override_settings

from hidp import config
from hidp.federated.oidc import jwks

from ..test_providers.example import ExampleOIDCClient


def _mock_response(content, *, status_code=200):
    response = requests.Response()
    response._content = content  # noqa: SLF001 (protected attribute)
    response.status_code = status_code
    return response


@override_settings(
    CACHES={
        "default": {
            # The module relies on caching to behave correctly.
            "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        }
    },
)
class TestJwksStore(TestCase):
    @classmethod
    def setUpTestData(cls):
        key = JWK(generate="RSA")
        key_set = JWKSet()
        key_set.add(key)
        cls.key_set = key_set.export(private_keys=False)

    def setUp(self):
        self.oidc_client = ExampleOIDCClient(client_id="test")
        config.configure_oidc_clients(self.oidc_client)
        cache.clear()

    def test_get_jwks_for_unregistered_client(self):
        """If the client is not registered, raise a KeyError."""
        with self.assertRaisesMessage(
            KeyError, "Client is not registered for 'example'."
        ):
            jwks.get_oidc_client_jwks(ExampleOIDCClient(client_id="test"))

    @mock.patch.object(jwks.requests, "get")
    def test_reluctantly_fetches_jwks_on_cache_miss(self, mock_get):
        """If signing keys are not cached, fetch them from the OIDC provider."""
        # Just raise an exception to stop the function early.
        # This test doubles as a test for request failure.
        mock_get.side_effect = [requests.RequestException("error")]
        with self.assertLogs(logger=jwks.logger, level="WARNING") as logs:
            jwks_data = jwks.get_oidc_client_jwks(self.oidc_client)

        # Complains about having to fetch signing keys
        self.assertEqual(
            logs.records[0].getMessage(),
            "Signing keys for 'example' are not cached,"
            " reluctantly fetching from 'https://example.com/jwks'.",
        )

        # Fetches signing keys from the OIDC provider
        mock_get.assert_called_once_with(
            self.oidc_client.jwks_uri,
            headers={"Accept": "application/json"},
            timeout=(5, 30),
        )

        # Logs the failure to fetch signing keys
        self.assertEqual(
            logs.records[1].getMessage(),
            "Failed to fetch signing keys for 'example' from 'https://example.com/jwks'.",
        )

        # Return None after failing to fetch signing keys
        self.assertIsNone(
            jwks_data, "Expected None after failing to fetch signing keys."
        )

        with self.subTest("Failure is cached"):
            # The second call doesn't try to fetch the data again.
            with self.assertNoLogs():
                self.assertIsNone(
                    jwks.get_oidc_client_jwks(self.oidc_client),
                    "Expected None after failing to fetch signing keys.",
                )
            mock_get.assert_called_once()  # No new requests are made

    @mock.patch.object(jwks.requests, "get")
    def test_handles_error_response(self, mock_get):
        """Log an exception when the JWKS endpoint returns an error response."""
        mock_get.return_value = _mock_response(b"Not Found", status_code=404)

        with self.assertLogs(logger=jwks.logger, level="ERROR") as logs:
            # This also triggers the cache miss warning, but it's not checked here,
            # as the previous test already covers that.
            jwks_data = jwks.get_oidc_client_jwks(self.oidc_client)

        mock_get.assert_called()  # The request is made

        self.assertIsNone(
            jwks_data, "Expected None after failing to fetch signing keys."
        )
        self.assertEqual(
            logs.records[0].getMessage(),
            "Error after fetching signing keys for 'example'"
            " from 'https://example.com/jwks': 404.",
        )

        with self.subTest("Failure is cached"):
            # The second call doesn't try to fetch the data again.
            with self.assertNoLogs():
                self.assertIsNone(
                    jwks.get_oidc_client_jwks(self.oidc_client),
                    "Expected None after failing to fetch signing keys.",
                )
            mock_get.assert_called_once()  # No new requests are made

    @mock.patch.object(jwks.requests, "get")
    def tests_handles_invalid_response(self, mock_get):
        """Log an exception when the JWKS endpoint returns an invalid response."""
        mock_get.return_value = _mock_response(
            b"This is not JSON, you must be mistaken."
        )

        with self.assertLogs(logger=jwks.logger, level="ERROR") as logs:
            jwks_data = jwks.get_oidc_client_jwks(self.oidc_client)

        self.assertIsNone(
            jwks_data, "Expected None after failing to fetch signing keys."
        )
        self.assertEqual(
            logs.records[0].getMessage(),
            "Failed to decode signing keys for 'example' from 'https://example.com/jwks'.",
        )

        with self.subTest("Failure is cached"):
            # The second call doesn't try to fetch the data again.
            with self.assertNoLogs():
                self.assertIsNone(
                    jwks.get_oidc_client_jwks(self.oidc_client),
                    "Expected None after failing to fetch signing keys.",
                )
            mock_get.assert_called_once()

    @mock.patch.object(jwks.requests, "get")
    def test_valid_response(self, mock_get):
        """Return the signing keys when the JWKS endpoint returns a valid response."""
        # This test doubles as a test for successful request caching,
        # and also tests eager fetching.
        mock_get.return_value = _mock_response(self.key_set.encode())

        jwk_data = jwks.get_oidc_client_jwks(self.oidc_client)

        mock_get.assert_called_once()  # The request is made
        self.assertIsInstance(jwk_data, JWKSet)

        with self.subTest("Success is cached"):
            # The second call doesn't try to fetch the data again.
            self.assertEqual(jwk_data, jwks.get_oidc_client_jwks(self.oidc_client))
            mock_get.assert_called_once()  # No new requests are made

        with self.subTest("Cache is bypassed"):
            # Cache is ignored, so two requests are made.
            self.assertEqual(
                jwk_data, jwks.get_oidc_client_jwks(self.oidc_client, eager=True)
            )
            self.assertEqual(
                len(mock_get.mock_calls), 2, "Expected two requests to be made."
            )

    @mock.patch.object(jwks.requests, "get")
    @mock.patch.object(jwks.cache, "get")
    def test_invalid_cache_value(self, mock_cache_get, mock_requests_get):
        """Log an exception when the cache contains an invalid value."""
        mock_cache_get.return_value = b"This is not JSON, you must be mistaken."
        mock_requests_get.return_value = _mock_response(self.key_set.encode())

        with self.assertLogs(logger=jwks.logger, level="ERROR") as logs:
            jwks_data = jwks.get_oidc_client_jwks(self.oidc_client)

        self.assertEqual(
            logs.records[0].getMessage(),
            "Failed to decode signing keys for 'example' from cache.",
        )

        # Falls back to fetching the data from the provider.
        mock_requests_get.assert_called_once()  # The request is made
        self.assertIsInstance(jwks_data, JWKSet)


@mock.patch.object(jwks, "get_oidc_client_jwks", autospec=True)
class RefreshJwks(SimpleTestCase):
    def setUp(self):
        class AnotherOIDCClient(ExampleOIDCClient):
            provider_key = "another"

        self.oidc_clients = [
            ExampleOIDCClient(client_id="test"),
            AnotherOIDCClient(client_id="another"),
        ]
        config.configure_oidc_clients(*self.oidc_clients)

    def test_refresh_registered_oidc_clients_jwks(self, mock_get_oidc_client_jwks):
        """Eagerly fetches the signing keys for all registered OIDC clients."""
        jwks.refresh_registered_oidc_clients_jwks()
        mock_get_oidc_client_jwks.assert_has_calls(
            [mock.call(client, eager=True) for client in self.oidc_clients]
        )

    def test_refresh_oidc_clients_jwks_management_command(
        self, mock_get_oidc_client_jwks
    ):
        """Eagerly fetches the signing keys and logs the process to stdout."""
        stdout = io.StringIO()
        call_command("refresh_oidc_clients_jwks", stdout=stdout)
        mock_get_oidc_client_jwks.assert_has_calls(
            [mock.call(client, eager=True) for client in self.oidc_clients]
        )
        self.assertEqual(
            [
                f"Fetching signing keys for '{client.provider_key}'"
                f" from '{client.jwks_uri}'..."
                for client in self.oidc_clients
            ],
            stdout.getvalue().splitlines(),
        )



from django_otp.plugins.otp_totp.models import TOTPDevice

from django.test import TestCase

from hidp.otp.devices import get_device_for_user, get_or_create_devices
from hidp.otp.exceptions import MultipleOtpDevicesError, NoOtpDeviceError
from hidp.test.factories import otp_factories, user_factories


class TestGetOrCreateDevices(TestCase):
    def test_get_or_create_devices(self):
        user = user_factories.UserFactory()
        totp_device, static_device = get_or_create_devices(user)
        self.assertEqual(totp_device.name, "Authenticator app")
        self.assertFalse(
            totp_device.confirmed, "Expected TOTP device to be unconfirmed"
        )
        self.assertEqual(static_device.name, "Recovery codes")
        self.assertFalse(
            static_device.confirmed, "Expected static device to be unconfirmed"
        )
        self.assertEqual(static_device.token_set.count(), 10)

    def test_get_or_create_devices_existing(self):
        user = user_factories.UserFactory()
        existing_totp_device = otp_factories.TOTPDeviceFactory(
            user=user, name="My existing authenticator app", confirmed=True
        )
        existing_static_device = otp_factories.StaticDeviceFactory(
            user=user, name="My existing backup tokens", confirmed=True
        )
        totp_device, static_device = get_or_create_devices(user)

        self.assertEqual(existing_totp_device, totp_device)
        self.assertEqual(existing_static_device, static_device)

        self.assertEqual(totp_device.name, "My existing authenticator app")
        self.assertTrue(totp_device.confirmed, "Expected TOTP device to be confirmed")
        self.assertEqual(static_device.name, "My existing backup tokens")
        self.assertTrue(
            static_device.confirmed, "Expected static device to be confirmed"
        )
        self.assertEqual(static_device.token_set.count(), 10)

    def test_static_tokens_are_not_replenished(self):
        user = user_factories.UserFactory()
        existing_static_device = otp_factories.StaticDeviceFactory(
            user=user, name="My existing backup tokens", confirmed=True
        )
        existing_static_device.token_set.create(token="old token")

        _totp_device, static_device = get_or_create_devices(user)
        self.assertEqual(static_device.token_set.count(), 1)
        self.assertEqual(static_device.token_set.first().token, "old token")


class TestGetDeviceForUser(TestCase):
    def test_get_device_for_user(self):
        user = user_factories.UserFactory()
        totp_device = otp_factories.TOTPDeviceFactory(user=user, confirmed=True)
        self.assertEqual(totp_device, get_device_for_user(user, TOTPDevice))

    def test_get_device_for_user_no_device(self):
        user = user_factories.UserFactory()
        with self.assertRaises(NoOtpDeviceError):
            get_device_for_user(user, TOTPDevice)

    def test_get_device_for_user_multiple_devices(self):
        user = user_factories.UserFactory()
        otp_factories.TOTPDeviceFactory(user=user, confirmed=True)
        otp_factories.TOTPDeviceFactory(user=user, confirmed=True)
        with self.assertRaises(MultipleOtpDevicesError):
            get_device_for_user(user, TOTPDevice)

    def test_get_device_for_user_with_only_unconfirmed_devices(self):
        user = user_factories.UserFactory()
        otp_factories.TOTPDeviceFactory(user=user, confirmed=False)
        with self.assertRaises(NoOtpDeviceError):
            get_device_for_user(user, TOTPDevice)



from django.test import SimpleTestCase

from hidp.federated.providers.base import OIDCClient

from .example import ExampleOIDCClient


class UnfinishedOIDCClient(OIDCClient):
    # Missing all the required attributes
    ...


class UnsafeOIDCClient(OIDCClient):
    # A seemingly valid OIDC client, but the endpoints do not utilize TLS.
    provider_key = "unsafe"
    issuer = "http://example.com"
    authorization_endpoint = "http://example.com/auth"
    token_endpoint = "http://example.com/token"
    userinfo_endpoint = "http://example.com/userinfo"
    jwks_uri = "http://example.com/jwks"


class BadIdentifierOIDCClient(ExampleOIDCClient):
    # A valid OIDC client, but with a provider key that is not URL-safe.
    provider_key = "Bad Identifier (test)"


class NamedExampleOIDCClient(ExampleOIDCClient):
    # A valid OIDC client, but with a custom name.
    name = "Named Example"


class NoPKCEOIDCClient(ExampleOIDCClient):
    # A valid OIDC client, but lacking PKCE support.
    has_pkce_support = False


class TestCustomProvider(SimpleTestCase):
    def test_invalid_provider(self):
        """Missing attributes raise a NotImplementedError."""
        with self.assertRaisesMessage(
            NotImplementedError,
            "'UnfinishedOIDCClient' misses (some of) the required attributes.",
        ):
            UnfinishedOIDCClient(client_id="test")

    def test_unsafe_endpoints(self):
        """An OIDC client with unsafe endpoints raises a ValueError."""
        with self.assertRaisesMessage(
            ValueError,
            "All endpoints must use TLS (https): 'http://example.com/auth' does not.",
        ):
            UnsafeOIDCClient(client_id="test")

    def test_bad_identifier(self):
        """A provider key that is not URL-safe raises a ValueError."""
        with self.assertRaisesMessage(
            ValueError,
            "'BadIdentifierOIDCClient.provider_key' is not URL-safe:"
            " 'Bad Identifier (test)'",
        ):
            BadIdentifierOIDCClient(client_id="test")

    def test_invalid_callback_base_url(self):
        """An invalid callback base URL raises a ValueError."""
        with self.assertRaisesMessage(
            ValueError,
            "Invalid callback base url: 'localhost:8000/example'."
            " Should be in the form of 'https://<netloc>'"
            " (path, querystring and/or fragment are not allowed).",
        ):
            ExampleOIDCClient(
                client_id="test",
                callback_base_url="localhost:8000/example",
            )

    def test_insecure_callback_base_url(self):
        """An insecure callback base URL raises a ValueError."""
        with self.assertRaisesMessage(
            ValueError,
            "Invalid callback base url: 'http://example.com/'."
            " Should be in the form of 'https://<netloc>'"
            " (path, querystring and/or fragment are not allowed).",
        ):
            ExampleOIDCClient(
                client_id="test",
                callback_base_url="http://example.com/",
            )

    def test_valid_callback_base_url(self):
        """A valid callback_base_url is stored."""
        client = ExampleOIDCClient(
            client_id="test", callback_base_url="https://example.com/"
        )
        self.assertEqual(client.client_id, "test")
        self.assertEqual(client.callback_base_url, "https://example.com/")

    def test_client_with_secret(self):
        """A client with a secret is stored."""
        client = ExampleOIDCClient(client_id="test", client_secret="secret")
        self.assertEqual(client.client_id, "test")
        self.assertEqual(client.client_secret, "secret")

    def test_client_without_pkce_requires_secret(self):
        """A client without PKCE support requires a client secret."""
        with self.assertRaisesMessage(
            ValueError,
            (
                "Please provide a client secret."
                " 'NoPKCEOIDCClient' declares it does not support PKCE,"
                " which means a client secret must be required for token exchange."
            ),
        ):
            NoPKCEOIDCClient(client_id="test")

    def test_minimal_client(self):
        """client_secret and callback_base_url default to None if not provided."""
        client = ExampleOIDCClient(client_id="test")
        self.assertEqual(client.client_id, "test")
        self.assertEqual(client.name, "Example")
        self.assertEqual(client.client_secret, None)
        self.assertEqual(client.callback_base_url, None)

    def test_get_issuer(self):
        """The issuer is returned by get_issuer."""
        client = ExampleOIDCClient(client_id="test")
        self.assertEqual(client.issuer, client.get_issuer(claims={}))

    def test_named_client(self):
        """Custom name takes precedence over the default name."""
        client = NamedExampleOIDCClient(client_id="test")
        self.assertEqual(client.name, "Named Example")



import uuid

from datetime import timedelta

from ..accounts.tokens import BaseTokenGenerator


class BaseOIDCTokenGenerator(BaseTokenGenerator):
    def make_token(self):
        """
        Generate an expiring token based on a random value.

        Returns:
            str: The generated token.
        """
        return super().make_token(str(uuid.uuid4()))

    def check_token(self, token):
        """
        Verify a token.

        Args:
            token (str): The token to verify.

        Returns:
            bool: True if the token is valid, False otherwise.
        """
        return super().check_token(token) is not None


class OIDCRegistrationTokenGenerator(BaseOIDCTokenGenerator):
    """Token for the OIDC registration process."""

    key_salt = "oidc-registration"
    token_timeout = timedelta(minutes=15).total_seconds()


class OIDCLoginTokenGenerator(BaseOIDCTokenGenerator):
    """Token for the OIDC login process."""

    key_salt = "oidc-login"
    token_timeout = timedelta(minutes=5).total_seconds()


class OIDCAccountLinkTokenGenerator(BaseOIDCTokenGenerator):
    """Token for the OIDC account linking process."""

    key_salt = "oidc-account-link"
    token_timeout = timedelta(minutes=15).total_seconds()



from django.apps import AppConfig


class HidpChecksMixin:
    def ready(self):  # noqa: PLR6301 (no-self-use)
        # registers checks
        from .config import checks  # noqa: F401, PLC0415


class HidpConfig(HidpChecksMixin, AppConfig):
    name = "hidp"



from unittest import mock

from django.core import mail
from django.test import TestCase, override_settings
from django.test.client import ClientHandler
from django.urls import reverse

from hidp.accounts import auth as hidp_auth
from hidp.accounts.forms import AuthenticationForm, RateLimitedAuthenticationForm
from hidp.test.factories import user_factories


class RateLimitedHandler(ClientHandler):
    def get_response(self, request):
        request.limited = True
        return super().get_response(request)


@override_settings(LOGIN_REDIRECT_URL="/", LANGUAGE_CODE="en")
class TestLogin(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()
        cls.login_url = reverse("hidp_accounts:login")

    def test_get_login(self):
        response = self.client.get(self.login_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "hidp/accounts/login.html")
        self.assertIn("form", response.context)
        self.assertIsInstance(response.context["form"], AuthenticationForm)

    @mock.patch("hidp.accounts.views.hidp_auth.login", wraps=hidp_auth.login)
    def test_valid_login_wrapped_login_function(self, mock_login):
        self.client.post(
            self.login_url,
            {
                "username": self.user.email,
                "password": "P@ssw0rd!",
            },
        )
        mock_login.assert_called_once()

    @mock.patch("hidp.accounts.views.hidp_auth.login", wraps=hidp_auth.login)
    def test_valid_login_unverified_email(self, mock_login):
        user = user_factories.UserFactory()
        with (
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_subject.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_body.txt"),
            self.assertTemplateUsed("hidp/accounts/verification/email/verification_body.html"),
        ):  # fmt: skip
            response = self.client.post(
                self.login_url,
                {
                    "username": user.email,
                    "password": "P@ssw0rd!",
                },
                follow=True,
            )
        # Does not log in the user
        mock_login.assert_not_called()
        # Verification email sent
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(
            message.subject,
            "Verify your email address",
        )
        # Redirected to verification required page
        self.assertRedirects(
            response,
            reverse(
                "hidp_accounts:email_verification_required", kwargs={"token": "email"}
            ),
        )
        # Verification required page
        self.assertInHTML(
            "Verification required",
            response.content.decode("utf-8"),
        )

    def test_valid_login_default_redirect(self):
        response = self.client.post(
            self.login_url,
            {
                "username": self.user.email,
                "password": "P@ssw0rd!",
            },
        )
        self.assertRedirects(response, "/", fetch_redirect_response=False)

    def test_valid_login_safe_next_param(self):
        response = self.client.post(
            self.login_url,
            {
                "username": self.user.email,
                "password": "P@ssw0rd!",
                "next": "/example/",
            },
        )
        self.assertRedirects(response, "/example/", fetch_redirect_response=False)

    def test_valid_login_unsafe_next_param(self):
        response = self.client.post(
            self.login_url,
            {
                "username": self.user.email,
                "password": "P@ssw0rd!",
                "next": "https://example.com/",
            },
        )
        self.assertRedirects(response, "/", fetch_redirect_response=False)

    def test_invalid_login(self):
        response = self.client.post(
            self.login_url,
            {
                "username": self.user.email,
                "password": "invalid",
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "hidp/accounts/login.html")
        self.assertIn("form", response.context)
        self.assertIsInstance(response.context["form"], AuthenticationForm)
        self.assertFormError(
            response.context["form"],
            None,
            (
                "Please enter a correct email address and password."
                " Note that both fields may be case-sensitive."
            ),
        )

    def test_rate_limited_login(self):
        self.client.handler = RateLimitedHandler(enforce_csrf_checks=False)
        response = self.client.post(
            self.login_url,
            {
                "username": self.user.email,
                "password": "P@ssw0rd!",
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "hidp/accounts/login.html")
        self.assertIn("form", response.context)
        self.assertIsInstance(response.context["form"], RateLimitedAuthenticationForm)
        self.assertFormError(
            response.context["form"],
            "i_am_not_a_robot",
            ("Please confirm that you are not a robot."),
        )

        response = self.client.post(
            self.login_url,
            {
                "username": self.user.email,
                "password": "P@ssw0rd!",
                "i_am_not_a_robot": "on",
            },
        )
        self.assertRedirects(response, "/", fetch_redirect_response=False)



from django_ratelimit.exceptions import Ratelimited
from django_ratelimit.middleware import RatelimitMiddleware as _RatelimitMiddleware

from ..rate_limit.views import rate_limited


class RateLimitMiddleware(_RatelimitMiddleware):
    rate_limited_view = rate_limited

    def process_exception(self, request, exception):
        if not isinstance(exception, Ratelimited):
            return None
        return self.rate_limited_view(exception)



from django import forms
from django.contrib.auth import forms as auth_forms
from django.contrib.auth import get_user_model
from django.contrib.auth import password_validation as django_password_validation
from django.db import transaction
from django.urls import reverse_lazy
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from .email_change import Recipient
from .models import EmailChangeRequest

UserModel = get_user_model()


class TermsOfServiceMixin:
    @staticmethod
    def create_agreed_to_tos_field():
        label = mark_safe(  # noqa: S308 (safe string, no user input)
            format_lazy(
                _('I have read and accept the <a href="{url}">Terms of Service</a>'),
                url=reverse_lazy("hidp_accounts:tos"),
            )
        )
        return forms.BooleanField(label=label, required=True)

    def set_agreed_to_tos(self, user):
        """Populate the `agreed_to_tos` field, if the user agreed."""
        if self.cleaned_data.get("agreed_to_tos", False):
            # Ensure the user has agreed to the terms of service.
            # Subclasses may remove the field, or make it optional.
            user.agreed_to_tos = timezone.now()


class UserCreationForm(TermsOfServiceMixin, auth_forms.BaseUserCreationForm):
    """
    Default UserCreationForm, allows user to register with username and password.

    The user **must** agree to the terms of service to register.

    The username field is mapped to `User.USERNAME_FIELD`. This makes it possible
    to change the username field to a different one, such as an email address.

    The user is asked to enter the password twice to avoid typos.
    The password is validated using the validators configured in
    `settings.AUTH_PASSWORD_VALIDATORS`.
    """

    template_name = "hidp/accounts/forms/user_creation_form.html"

    # Fields
    agreed_to_tos = TermsOfServiceMixin.create_agreed_to_tos_field()
    # Remove the option to create an account with an unusable password.
    usable_password = None

    class Meta:
        model = UserModel
        fields = (UserModel.USERNAME_FIELD,)

    def _get_validation_exclusions(self):
        # Exclude email from model validation (unique constraint),
        # This will make the form valid even if the email is already in use.
        # This results in a IntegrityError when saving the user, which
        # must be handled by the view, to prevent user enumeration attacks.
        return {"email", *super()._get_validation_exclusions()}

    def save(self, *, commit=True):
        user = super().save(commit=False)
        self.set_agreed_to_tos(user)
        if commit:
            user.save()
        return user


class EmailVerificationForm(forms.ModelForm):
    """Store the date and time when the user verified their email address."""

    template_name = "hidp/accounts/verification/forms/email_verification_form.html"

    class Meta:
        model = UserModel
        fields = ["first_name", "last_name"]

    def __init__(self, *args, **kwargs):
        """
        Initialize the form for the given user.

        Remove the first and last name fields if they are both already filled in. This
        is possible when the user is created using an OIDC connection, and the given
        name and family name are provided by the OIDC provider.

        If they are not filled in, they are required.
        """
        super().__init__(*args, **kwargs)

        if self.instance.first_name and self.instance.last_name:
            self.fields.pop("first_name")
            self.fields.pop("last_name")
        else:
            self.fields["first_name"].required = True
            self.fields["last_name"].required = True

    def save(self, *, commit=True):
        """
        Mark the user as verified.

        Args:
            commit:
                Whether to save the user to the database after
                marking the user as verified.

        Returns:
            The user with the email address verified.
        """
        self.instance.email_verified = timezone.now()
        return super().save(commit=commit)


class AuthenticationForm(auth_forms.AuthenticationForm):
    """
    Default AuthenticationForm, allows user to log in with username and password.

    The username field is mapped to `User.USERNAME_FIELD`. This makes it possible
    to change the username field to a different one, such as an email address.
    """

    username = forms.EmailField(
        label=_("Email"),
        max_length=254,
        widget=forms.EmailInput(attrs={"autocomplete": "email", "autofocus": True}),
        required=True,
    )
    template_name = "hidp/accounts/forms/authentication_form.html"

    def __init__(self, request=None, *args, **kwargs):
        """
        Initialize the form with the given `request`.

        The `request` is stored in an instance variable, to allow all
        form methods to access the request.
        """
        super().__init__(request, *args, **kwargs)

    def is_valid(self):
        """
        Validate the username and password.

        Returns `True` if the credentials are valid and the user
        is allowed to log in, otherwise `False`.

        Validation errors are stored in the form's `errors` attribute.
        """
        return super().is_valid()

    def get_user(self):
        """
        Return the user authenticated by the form (after calling `is_valid`).

        Returns `None` if no user was authenticated.
        """
        return super().get_user()

    def get_invalid_login_error(self):
        """
        Hook to alter the error message when authentication fails.

        The default implementation returns a fixed message, regardless of
        the credentials provided by the user. This message is parameterized
        to use the name of the username field, as defined by the user model.

        To customize the error message, subclass this form and override
        `AuthenticationForm.messages['invalid_login']`.
        """
        return super().get_invalid_login_error()

    def confirm_login_allowed(self, user):
        """
        Hook to perform additional checks on the user, before logging them in.

        The default implementation checks if the user is active, and raises
        a `ValidationError` if the user is not active.

        To change the message raised when a user is active, subclass this form
        and override `AuthenticationForm.messages['inactive']`.

        Note:
        The default backend (`django.contrib.auth.backends.ModelBackend`) does
        not authenticate inactive users, and will not call this method for
        inactive users.

        To allow inactive users to authenticate, but prevent them from
        logging in, set `settings.AUTHENTICATION_BACKENDS` to
        `django.contrib.auth.backends.AllowAllUsersModelBackend`.
        """
        return super().confirm_login_allowed(user)


class PasswordResetRequestForm(forms.Form):
    """Start the password reset process by requesting a password reset email."""

    template_name = "hidp/accounts/recovery/forms/password_reset_request_form.html"

    # Fields
    email = forms.EmailField(
        label=_("Email"),
        max_length=254,
        widget=forms.EmailInput(attrs={"autocomplete": "email"}),
    )

    def get_user(self):
        """
        Given an email, return the user who should receive a reset.

        Returns None if no user is found, or the user is not allowed
        to reset their password (e.g. inactive).
        """
        return UserModel.objects.filter(
            email__iexact=self.cleaned_data["email"], is_active=True
        ).first()


class PasswordResetForm(auth_forms.SetPasswordForm):
    """
    Allows the user to set a new password without entering the old password.

    The user is asked to enter a new password twice to avoid typos.
    The password is validated using the validators configured in
    `settings.AUTH_PASSWORD_VALIDATORS`.
    """

    template_name = "hidp/accounts/recovery/forms/password_reset_form.html"

    def __init__(self, user, *args, **kwargs):
        """
        Initialize the form with the given `user`.

        The `user` is stored in an instance variable, to allow all
        form methods to access the user.
        """
        super().__init__(user, *args, **kwargs)

    def save(self, *, commit=True):
        """
        Save the new password for the user.

        Args:
            commit:
                Whether to save the user to the database after
                setting the password.

        Returns:
            The user with the new password set.
        """
        return super().save(commit=commit)


class PasswordChangeForm(auth_forms.PasswordChangeForm):
    """
    Allows the user to change their password by entering the old password.

    The user is asked to enter the old password, and a new password twice
    to avoid typos. The old password is used to verify the user's identity.
    The new password is validated using the validators configured in
    `settings.AUTH_PASSWORD_VALIDATORS`.
    """

    template_name = "hidp/accounts/management/forms/password_change_form.html"

    def __init__(self, user, *args, **kwargs):
        """
        Initialize the form with the given `user`.

        The `user` is stored in an instance variable, to allow all
        form methods to access the user.
        """
        super().__init__(user, *args, **kwargs)

    def save(self, *, commit=True):
        """
        Save the new password for the user.

        Args:
            commit:
                Whether to save the user to the database after
                setting the password.

        Returns:
            The user with the new password set.
        """
        return super().save(commit=commit)


class SetPasswordForm(auth_forms.SetPasswordForm):
    """Form for setting a new password without requiring the old password."""

    template_name = "hidp/accounts/management/forms/set_password_form.html"

    # Fields
    new_password1 = forms.CharField(
        label=_("Password"),
        required=True,
        strip=False,
        widget=forms.PasswordInput(attrs={"autocomplete": "new-password"}),
        help_text=django_password_validation.password_validators_help_text_html(),
    )
    new_password2 = forms.CharField(
        label=_("Password confirmation"),
        required=True,
        widget=forms.PasswordInput(attrs={"autocomplete": "new-password"}),
        strip=False,
        help_text=_("Enter the same password as before, for verification."),
    )


class RateLimitedAuthenticationForm(AuthenticationForm):
    """
    Authentication form that is used when a user is rate limited.

    This form includes a simple "I am not a robot" checkbox to demonstrate
    how additional protection can be added to an authentication form.

    It is recommended to replace this form with a more robust implementation
    that provides stronger protection against automated attacks.
    """

    template_name = "hidp/accounts/forms/rate_limited_authentication_form.html"

    # Fields
    i_am_not_a_robot = forms.BooleanField(
        label=_("I am not a robot"),
        required=True,
        error_messages={
            "required": _("Please confirm that you are not a robot."),
        },
    )


class EditUserForm(forms.ModelForm):
    template_name = "hidp/accounts/management/forms/edit_user_form.html"

    class Meta:
        model = UserModel
        fields = ("first_name", "last_name")


class EmailChangeRequestForm(forms.ModelForm):
    """
    Initiate the email address change flow for a user.

    The user is asked to enter their password to confirm their identity.
    """

    template_name = "hidp/accounts/management/forms/email_change_request_form.html"

    # Fields
    proposed_email = forms.EmailField(
        label=_("New email"),
        max_length=254,
        widget=forms.EmailInput(),
        help_text=_(
            "Please note that this also changes the username you use to sign in."
        ),
    )
    password = forms.CharField(
        label=_("Password"),
        strip=False,
        widget=forms.PasswordInput(attrs={"autocomplete": "current-password"}),
        help_text=_("Your password is required to verify your identity."),
    )

    class Meta:
        model = EmailChangeRequest
        fields = ["proposed_email"]

    def __init__(self, *args, user, **kwargs):
        """
        Initialize the form for the given user.

        The `user` is stored in an instance variable, to allow all
        form methods to access the user.
        """
        super().__init__(*args, **kwargs)
        self.user = user

    def clean_password(self):
        """
        Validate the password.

        Returns the password if it is correct, otherwise raises a `ValidationError`.
        """
        password = self.cleaned_data["password"]
        if not self.user.check_password(password):
            raise forms.ValidationError(_("The password is incorrect."))
        return password

    def clean_proposed_email(self):
        """
        Validate the proposed email address.

        Returns the proposed email address if it is different from the current email
        address of the user, otherwise raises a `ValidationError`.
        """
        proposed_email = self.cleaned_data["proposed_email"]
        if proposed_email == self.user.email:
            raise forms.ValidationError(
                _("The new email address is the same as the current email address.")
            )
        return proposed_email

    def save(self, *, commit=True):
        """
        Create an email change request.

        Replaces any existing email change requests for the user.

        Args:
            commit:
                Whether to save the email change request to the database
                after creating it.

        Returns:
            The email change request.
        """
        # Set the user on the email change request.
        instance = super().save(commit=False)
        instance.user = self.user
        instance.current_email = self.user.email

        if commit:
            with transaction.atomic():
                # Remove existing email change requests for the user, if any.
                EmailChangeRequest.objects.filter(user=self.user).delete()
                instance.save()
        return instance


class EmailChangeConfirmForm(forms.ModelForm):
    """Update the EmailChangeRequest with the correct confirmation."""

    template_name = "hidp/accounts/management/forms/email_change_confirm_form.html"

    # Fields
    allow_change = forms.BooleanField(
        label=_("Yes, I want to change my email address"),
        required=True,
    )

    class Meta:
        model = EmailChangeRequest
        fields = []

    def __init__(self, *args, recipient, **kwargs):
        """Initialize the form for the given email change request."""
        self.recipient = recipient
        super().__init__(*args, **kwargs)

    def save(self, *, commit=True, **kwargs):
        instance = super().save(commit=False)

        match self.recipient:
            case Recipient.CURRENT_EMAIL:
                instance.confirmed_by_current_email = True
            case Recipient.PROPOSED_EMAIL:
                instance.confirmed_by_proposed_email = True
            case _:
                raise ValueError("Invalid recipient")

        # Change email address of user if complete.
        if commit:
            with transaction.atomic():
                instance.save()
                if instance.is_complete():
                    instance.user.email = instance.proposed_email
                    instance.user.save(update_fields=["email"])

        return instance


class EmailChangeCancelForm(forms.Form):
    """Delete the EmailChangeRequest."""

    template_name = "hidp/accounts/management/forms/email_change_cancel_form.html"

    # Fields
    allow_cancel = forms.BooleanField(
        label=_("Yes, I want to cancel changing my email address"),
        required=True,
        initial=True,
        widget=forms.HiddenInput(),
    )
